import functools

import jax
from jax import numpy as jnp

# def itertangents(args, argnum):
#     for i, x in enumerate(args):
#         if i == argnum:
#             yield jnp.ones(shape, dtype)
#         else:
#             yield jnp.zeros(shape, dtype)

def elementwise_grad(fun, argnum=0):
    assert int(argnum) == argnum and argnum >= 0, argnum
    @functools.wraps(fun)
    def funderiv(*args):
        preargs = args[:argnum]
        postargs = args[argnum + 1:]
        def oneargfun(arg):
            args = preargs + (arg,) + postargs
            return fun(*args)
        primal = args[argnum]
        shape = getattr(primal, 'shape', ())
        dtype = getattr(primal, 'dtype', type(primal))
        tangent = jnp.ones(shape, dtype)
        primal_out, tangent_out = jax.jvp(oneargfun, (primal,), (tangent,))
        return tangent_out
    return funderiv
    
x = jnp.arange(6.).reshape(2, 3)
fun = jnp.sin

print(jax.vmap(jax.vmap(jax.grad(fun), 0), 1)(x.T))
print(elementwise_grad(fun)(x))

print(jax.vmap(jax.vmap(jax.grad(jax.grad(fun)), 0), 1)(x.T))
print(elementwise_grad(elementwise_grad(fun))(x))

f = lambda x, y: 2 * x + 3 * y
print(elementwise_grad(f)(0., 0.))
print(elementwise_grad(f, 1)(0., 0.))
print(elementwise_grad(f)(jnp.zeros((1, 5)), jnp.zeros((3, 1))))



import jax
from jax import tree_util

@tree_util.register_pytree_node_class
class cippa:
    
    def __init__(self, x):
        self.x = x
    
    def __repr__(self):
        return repr(self.x)
    
    def tree_flatten(self):
        return (self.x,), None
    
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        self = cls.__new__(cls)
        self.x, = children
        return self

class lippa(cippa): pass

print(jax.jacfwd(lambda x: x)(cippa(0.)))
print(jax.jacfwd(lambda x: x)(lippa(0.))) # -> error, not registered



import jax

@jax.jit
def f(a, b):
    print('f: a =', a)
    print('f: b =', b)
    @jax.jit
    def g(a):
        print('g: a =', a)
        print('g: b =', b)
        return b * a
    return g(a)

print(f(2, 3))
print(f(2, 5))



import jax
from jax import numpy as jnp
import numpy as np

def check_aux(aux):
    print()
    for op in [jax.grad, jax.jacrev, jax.jacfwd]:
        print(op.__name__)
        print(op(lambda x: (x, aux), has_aux=True)(0.))

@jax.tree_util.register_pytree_node_class
class cippa:

    def __init__(self, **kw):
        self.kw = kw
        self.names = None

    def tree_flatten(self):
        if self.names is None:
            self.names = [n for n, x in self.kw.items() if isinstance(x, jnp.ndarray)]
        children = tuple(self.kw[n] for n in self.names)
        return children, self.names

    @classmethod
    def tree_unflatten(cls, names, children):
        self = cls(**dict(zip(names, children)))
        self.names = names

check_aux(cippa(a='ciao'))
check_aux(cippa(a=jnp.zeros(1)))



import functools

import jax
from jax import numpy as jnp

class Cippa:
    pass

def oldinit(self, x):
    self.x = x
    
# @functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
# def init(self, x):
#     oldinit(self, x)
#     return 0.
#
# @init.defjvp
# def initjvp(self, primals, tangents):
#     x, = primals
#     x_dot, = tangents
#     return init(self, x), x_dot

def notracer(x):
    if isinstance(x, jax.core.Tracer):
        return notracer(x.primal)
    else:
        return x
    
def __init__(self, x):
    oldinit(self, notracer(x))
    self._x = x
    
def oldciao(self, a):
    return self.x * a

@functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
def ciaoimpl(self, x, a):
    return oldciao(self, a)

@ciaoimpl.defjvp
def ciaojvp(self, primals, tangents):
    x, a = primals
    x_dot, a_dot = tangents
    return ciaoimpl(self, x, a), x_dot * a + x * a_dot

def ciao(self, a):
    return ciaoimpl(self, self._x, a)

Cippa.__init__ = __init__
Cippa.ciao = ciao

@functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
def f(a, x):
    print(a)
    return x

@f.defjvp
def fjvp(a, primals, tangents):
    x, = primals
    x_dot, = tangents
    return f(a, x), x_dot

jax.grad(lambda x: Cippa(x).ciao(2.))(3.)
# def g(x):
#     return f('cippa', x)
# print(jax.grad(g)(2.))



import numpy as np
from matplotlib import pyplot as plt

def f(d):
    x = d % (2 * np.pi)
    return np.minimum(x, 2 * np.pi - x)

fig, ax = plt.subplots(num='circular', clear=True)

x = np.linspace(-20, 20, 1000)
y = f(x)

ax.plot(x, y)

fig.show()



# lsqfitgp/docs/conf.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'lsqfitgp'
author = 'Giacomo Petrillo'

from datetime import datetime
now = datetime.now()
year = '2020'
if now.year > int(year):
    year += '-' + str(now.year)
copyright = year + ', ' + author


# # The full version, including alpha/beta/rc tags
import lsqfitgp
release = lsqfitgp.__version__
version = release
project += ' ' + version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # 'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'numpydoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

html_title = f'{project} documentation'

html_theme_options = dict(
    description = 'A general purpose Gaussian process regression module',
    fixed_sidebar = True,
    github_button = True,
    github_type = 'star',
    github_repo = 'lsqfitgp',
    github_user = 'Gattocrucco',
    show_relbars = True,
)

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

master_doc = 'index'


# -- Other options -------------------------------------------------

autoclass_content = 'both'
autodoc_preserve_defaults = True

numpydoc_class_members_toctree = False
numpydoc_show_class_members = False



# lsqfitgp/tests/test_copula_invgamma.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" Test the copula.invgamma module """

import sys

from jax import test_util
from scipy import stats
import pytest
import numpy as np

sys.path.insert(0, '.')
from lsqfitgp.copula import _invgamma

def test_grad():
    test_util.check_grads(_invgamma.gammainccinv, (2.5, 0.3), 1)
    test_util.check_grads(_invgamma.gammaincinv, (2.5, 0.3), 1)

@pytest.mark.xfail
def test_hess():
    test_util.check_grads(_invgamma.gammainccinv, (2.5, 0.3), 2)
    test_util.check_grads(_invgamma.gammaincinv, (2.5, 0.3), 2)

def test_ppf():
    q = 0.43
    a = 3.6
    b = 2.1
    np.testing.assert_array_max_ulp(
        stats.invgamma.ppf(q, a, scale=b),
        _invgamma.invgamma.ppf(q, a, scale=b),
    )

def test_isf():
    q = 0.43
    a = 3.6
    b = 2.1
    np.testing.assert_allclose(
        _invgamma.invgamma.isf(q, a, scale=b),
        stats.invgamma.isf(q, a, scale=b),
        atol=0, rtol=1e-7,
    )

def test_logpdf():
    q = 0.43
    a = 3.6
    b = 2.1
    np.testing.assert_allclose(
        _invgamma.invgamma.logpdf(q, a, scale=b),
        stats.invgamma.logpdf(q, a, scale=b), 
        rtol=1e-5,
    )

def test_cdf():
    q = 0.43
    a = 3.6
    b = 2.1
    np.testing.assert_allclose(
        _invgamma.invgamma.cdf(q, a, scale=b),
        stats.invgamma.cdf(q, a, scale=b), 
        rtol=1e-6,
    )



"""Fit of parton distributions functions (PDFs)

Like pdf2, but with correct integral constraints and naming this time"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar

np.random.seed(20220416)

#### COMPONENTS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

pid  = flavor['f0']
name = flavor['f1']

nflav = len(flavor)
nx    = 30
ndata = 20

indices = dict(
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

#### MODEL ####
# for each component of the proton:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='pid')

xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

M = np.random.randn(ndata, nflav, nx) # transformation PDF(X) -> data

xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### CREATE GP OBJECT ####

gp = lgp.GP()

gp.addproc(kernel, 'h')
gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
gp.addproctransf({
    'primitive': lambda x: x['x'],
    'h'        : -1,
}, 'primitive of xf(x)')

# data
gp.addx(xdata, 'xdata', proc='f')
gp.addtransf({'xdata': M}, 'data', axes=2)

# total momentum rule
gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)

# quark sum rules
qdiff = np.array([1, -1])[:, None]
for quark in 'ducs':
    idx = indices[quark]
    label = f'{quark}{quark}bar' # the one appearing in `constraints`
    xlabel = f'x{label}'
    gp.addx(xinteg[idx], xlabel, proc='primitive')
    gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)

#### GENERATE FAKE DATA ####

prior = gp.predfromdata(constraints, ['data', 'xdata'])
priorsample = next(gvar.raniter(prior))

datamean = priorsample['data']
dataerr = np.full_like(datamean, 1)
datamean = datamean + dataerr * np.random.randn(*dataerr.shape)
data = gvar.gvar(datamean, dataerr)

# check sum rules approximately with trapezoid rule
def check_integrals(x, y):
    checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
    print('sum_i int dx x f_i(x) =', checksum)
    for q in 'ducs':
        idx = indices[q]
        qx = x[idx]
        qy = y[idx]
        checksum = np.sum(qdiff * (qy[:, 1:] + qy[:, :-1]) / 2 * np.diff(qx, axis=1))
        print(f'sum_i={q}{q}bar int dx f_i(x) =', checksum)

print('check integrals in fake data:')
check_integrals(xdata['x'], priorsample['xdata'])

#### FIT ####

pred = gp.predfromdata(dict(**constraints, **{
    'data': data,
}), ['data', 'xdata'])

print('check integrals in fit:')
check_integrals(xdata['x'], pred['xdata'])

#### PLOT RESULTS ####

fig, axs = plt.subplots(1, 2, num='pdf3', clear=True, figsize=[9, 4.5])
axs[0].set_title('PDFs')
axs[1].set_title('Data')

for i in range(nflav):
    
    x = xdata[i]['x']
    ypdf = pred['xdata'][i]
    ydata = priorsample['xdata'][i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    
    color = 'C' + str(i // 2)
    if i % 2:
        kw = dict(hatch='//////', edgecolor=color, facecolor='none')
        kwp = dict(color=color, linestyle='--')
    else:
        kw = dict(alpha=0.6, facecolor=color)
        kwp = dict(color=color)
    
    axs[0].fill_between(x, m - s, m + s, label=name[i], **kw)
    axs[0].plot(x, ydata, **kwp)

axs[0].legend(fontsize='small')

m = gvar.mean(pred['data'])
s = gvar.sdev(pred['data'])
x = np.arange(len(data))
axs[1].fill_between(x, m - s, m + s, step='mid', color='lightgray')
axs[1].errorbar(x, datamean, dataerr, color='black', linestyle='', capsize=2)

fig.tight_layout()
fig.show()



import functools

import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial
from matplotlib import pyplot as plt

from lsqfitgp import _patch_jax

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1, 2, 3))
def taylor(coefgen, args, n, m, x):
    c = coefgen(n, n + m, *args)
    k = jnp.arange(n, n + m)
    c = c * jnp.exp(jspecial.gammaln(1 + k) - jspecial.gammaln(1 + k - n))
    return jnp.polyval(c[::-1], x)

@taylor.defjvp
def taylor_jvp(coefgen, args, n, m, primals, tangents):
    x, = primals
    xt, = tangents
    return taylor(coefgen, args, n, m, x), taylor(coefgen, args, n + 1, m, x) * xt

def sgngamma(x):
    return jnp.where((x > 0) | (x % 2 < 1), 1, -1)

def coefgen_iv(s, e, nu):
    m = jnp.arange(s, e)
    u = 1 + m + nu
    return sgngamma(u) * jnp.exp(-jspecial.gammaln(1 + m) - jspecial.gammaln(u))
    
def ivmod(nu, x):
    return taylor(coefgen_iv, (nu,), 0, 5, (x / 2) ** 2)

def kvmod(nu, x):
    factor = jnp.pi / (2 * jnp.sin(jnp.pi * nu))
    return factor * (ivmod(-nu, x) - (x / 2) ** (2 * nu) * ivmod(nu, x))

def kvmod2(nu, x):
    return (x / 2) ** nu * _patch_jax.kv(nu, x)

def ivmodx2_nearzero(nu, x2):
    return taylor(coefgen_iv, (nu,), 0, 5, x2 / 4)

def kvmodx2_nearzero(nu, x2):
    factor = jnp.pi / (2 * jnp.sin(jnp.pi * nu))
    return factor * (ivmodx2_nearzero(-nu, x2) - (x2 / 4) ** nu * ivmodx2_nearzero(nu, x2))

def kvmodx2(nu, x2):
    assert int(nu) != nu, nu
    nearzero = kvmodx2_nearzero(nu, x2)
    x = jnp.sqrt(x2)
    normal = (x / 2) ** nu * _patch_jax.kv(nu, x)
    return jnp.where(jnp.abs(x) < 1e-2, nearzero, normal)

fig, ax = plt.subplots(num='taylor5', clear=True)

def deriv(f, i, a):
    if i == 0:
        return f
    else:
        return deriv(jax.grad(f, a), i - 1, a)
        
def vec1(f):
    return jax.vmap(f, (None, 0), 0)

x = 0.01 * jnp.linspace(0, 1, 1000)
nu = 2.5

for n in range(5):
    f1 = deriv(kvmod2, n, 1)
    f2 = deriv(lambda v, x: kvmodx2(v, x ** 2), n, 1)
    f1 = vec1(f1)
    f2 = vec1(f2)
    y1 = f1(nu, x)
    y2 = f2(nu, x)
    ax.plot(x, y1, color=f'C{n}', label=f'{n}')
    ax.plot(x, y2, color=f'C{n}', linestyle='--')

ax.set_ylim(-1, 3)

fig.show()



# lsqfitgp/examples/h.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE H.

    Where at first sight nothing has changed, but behind the scenes
    important information has been lost forever.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

xdata = np.linspace(0, 10, 10)
xpred = np.linspace(-15, 25, 300)
y = np.sin(xdata)

gp = lgp.GP(lgp.ExpQuad(scale=3))
gp.addx(xdata, 'pere')
gp.addx(xpred, 'banane')

u = gp.predfromdata({'pere': y}, 'banane', keepcorr=False)
m = gvar.mean(u)
s = gvar.sdev(u)
cov = gvar.evalcov(u)

fig, ax = plt.subplots(num='h', clear=True)

patch = ax.fill_between(xpred, m - s, m + s, label='pred', alpha=0.5)
color = patch.get_facecolor()[0]
simulated_lines = np.random.multivariate_normal(m, cov, size=10)
ax.plot(xpred, simulated_lines.T, '-', color=color)
ax.plot(xdata, y, 'k.', label='data')
ax.legend(loc='best')

fig.show()



import numpy as np
from matplotlib import pyplot as plt

import lsqfitgp as lgp

fig, ax = plt.subplots(num='color', clear=True)

s = 0
x = np.linspace(s, s + 100, 10000)
for n in range(2, 30, 3):
    y = lgp.Color(n=n)(0, x)
    ax.plot(x, y, label=f'$n = {n}$')
ax.legend()

fig.show()



# lsqfitgp/_kernels/_zeta.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from jax import numpy as jnp

from .. import _special
from .. import _patch_jax
from .. import _Kernel
from .._Kernel import stationarykernel

def _zeta_derivable(nu=None):
    with _patch_jax.skipifabstract():
        return max(0, jnp.ceil(nu) - 1)

@stationarykernel(maxdim=1, derivable=_zeta_derivable, saveargs=True)
def _ZetaBase(delta, nu=None):
    """
    Zeta kernel.
    
    .. math::
        k(\\Delta)
        &= \\frac{\\Re F(\\Delta, s)}{\\zeta(s)} =
        \\qquad (s = 1 + 2 \\nu, \\quad \\nu \\ge 0) \\\\
        &= \\frac1{\\zeta(s)} \\sum_{k=1}^\\infty
        \\frac {\\cos(2\\pi k\\Delta)} {k^s} = \\\\
        &= -(-1)^{s/2}
        \\frac {(2\\pi)^s} {2s!}
        \\frac {\\tilde B_s(\\Delta)} {\\zeta(s)}
        \\quad \\text{for even integer $s$.}
    
    It is equivalent to fitting with a Fourier series of period 1 with
    independent priors on the coefficients with mean zero and variance
    :math:`1/(\\zeta(s)k^s)` for the :math:`k`-th term. Analogously to
    :class:`Matern`, the process is :math:`\\lceil\\nu\\rceil - 1` times
    derivable, and the highest derivative is continuous iff :math:`\\nu\\bmod 1
    \\ge 1/2`.
    
    Note that the :math:`k = 0` term is not included in the summation, so the
    mean of the process over one period is forced to be zero.
    
    Reference: Petrillo (2022).
    
    """
    
    # TODO reference as covariance function? I had found something with
    # fourier and bernoulli but lost it. Maybe known as zeta?
    
    # TODO add constant as option, otherwise I can't compute the Fourier
    # series when I add a constant. => Maybe this will be solved when I
    # improve the transformations system.
    
    # TODO ND version. The separable product is not equivalent I think.
    
    # TODO the derivative w.r.t. nu is probably broken
    
    with _patch_jax.skipifabstract():
        assert 0 <= nu < jnp.inf, nu
        
    s = 1 + 2 * nu
    nupos = _special.periodic_zeta(delta, s) / _special.zeta(s)
    nuzero = jnp.where(delta % 1, 0, 1)
    return jnp.where(s > 1, nupos, nuzero)
    
    # return -(-1) ** (s // 2) * _special.scaled_periodic_bernoulli(s, delta) / jspecial.zeta(s, 1)
    
    # TODO use the bernoully version for integer even s, based on the type of
    # the input so that it's static, because it is much more accurate
    
class Zeta(_ZetaBase):
    
    __doc__ = _ZetaBase.__doc__
    
    # TODO write a method of _KernelBase that makes a new kernel from the
    # current one to be used as starting point by all the transformation
    # methods. It should have an option on how much the subclass should be
    # preserved, for example this implementation of `fourier` is broken as soon
    # as a transformation is applied to the kernel. Linear transformations
    # should transform not only the kernel but also its transformations. I
    # have to think how to make this work in full generality. => Tentative
    # design: loc, scale, dim, etc. must become standalone methods and
    # appropriately transform the other transformation methods. Alternative:
    # interface for providing the internals of the transformations, separately
    # for cross and symmetric cases, and it's always the Kernel method that
    # manages everything like it is for __call__.
    
    def fourier(self, dox, doy):
        
        # TODO problem: this ignores completely loc and scale. Write a static
        # _KernelBase method that applies loc and scale to a kernel core, use
        # it in __init__ and here.
        
        if not dox and not doy:
            return self
        
        nu = self.initargs['nu']
        s = 1 + 2 * nu
        
        if dox and doy:
            def kernel(k, q):
                order = jnp.ceil(k / 2)
                denom = order ** s * _special.zeta(s)
                return jnp.where((k == q) & (k > 0), 1 / denom, 0)
        
        else:
            def kernel(k, y):
                order = jnp.ceil(k / 2)
                denom = order ** s * _special.zeta(s)
                odd = k % 2
                arg = 2 * jnp.pi * order * y
                return jnp.where(k > 0, jnp.where(odd, jnp.sin(arg), jnp.cos(arg)) / denom, 0)
        
            if doy:
                kernel = lambda x, q, kernel=kernel: kernel(q, x)
        
        cls = _Kernel.Kernel if dox == doy and isinstance(self, _Kernel.Kernel) else _Kernel.CrossKernel
        obj = cls(kernel)
        obj.initargs = self.initargs
        obj._maxderivable = self._maxderivable
        return obj



import itertools
import functools

import jax
from jax import numpy as jnp

class logfunc:
    
    _indent = 0
    
    def __init__(self, func):
        self._func = func
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kw):
        repr_args = map(repr, args)
        repr_kw = (f'{k}={v!r}' for k, v in kw.items())
        arglist = ', '.join(itertools.chain(repr_args, repr_kw))
        arglist = '...'
        print(4 * self._indent * ' ' + self._func.__name__ + '(' + arglist + ')')
        self.__class__._indent += 1
        try:
            result = self._func(*args, **kw)
        finally:
            self.__class__._indent -= 1
        return result

logfunc = lambda x: x

@jax.custom_jvp
@logfunc
def stop_hessian(x):
    return x

@stop_hessian.defjvp
@logfunc
def stop_hessian_jvp(p, t):
    x, = p
    xt, = t
    return x, stop_hessian_helper(x) * xt

@jax.custom_jvp
@logfunc
def stop_hessian_helper(x):
    return 1

@stop_hessian_helper.defjvp
@logfunc
def stop_hessian_helper_jvp(p, t):
    x, = p
    xt, = t
    return 0., xt

@logfunc
def f(x):
    return jnp.sin(jnp.cos(x))

@logfunc
def g(x):
    return 1/2 * x ** 2

@logfunc
def h(x):
    return g(stop_hessian(f(x)))

# h = 1/2 f^2
# h' = f f'
# h'' = f' f' + f f''
#               ^^^^^ stop_hessian should remove this

x = jnp.arange(4.)
f1 = jax.vmap(jax.grad(f))
f2 = jax.vmap(jax.grad(jax.grad(f)))

print("f' =", f1(x))
print('fwd:', jax.vmap(jax.jacfwd(lambda x: stop_hessian(f(x))))(x))
print('rev:', jax.vmap(jax.jacrev(lambda x: stop_hessian(f(x))))(x))
print()

print("f f' =", f(x) * f1(x))
print('fwd:  ', jax.vmap(jax.jacfwd(h))(x))
print('rev:  ', jax.vmap(jax.jacrev(h))(x))
print()

print("f' f' =    ", f1(x) ** 2)
print("f f'' =    ", f(x) * f2(x))
print("f'f'+ff'' =", f1(x) ** 2 + f(x) * f2(x))
print('fwd-fwd:   ', jax.vmap(jax.jacfwd(jax.jacfwd(h)))(x)) # ok -> f' f'
print('rev-fwd:   ', jax.vmap(jax.jacrev(jax.jacfwd(h)))(x)) # ok -> f' f'
print('fwd-rev:   ', jax.vmap(jax.jacfwd(jax.jacrev(h)))(x)) # wrong -> f f'' (?)
print('rev-rev:   ', jax.vmap(jax.jacrev(jax.jacrev(h)))(x)) # wrong -> f f'' (?)



"""Fit of parton distributions functions (PDFs)

The difference from pdf1.py is that we define the transformation on processes
instead of on their finite realizations"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar

np.random.seed(20220416)

#### DEFINE MODEL ####
# for each gluon:
# h ~ GP
# f = h''
# int_0^1 dx f(x) = [h'(x)]_0^1
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1

xtype = np.dtype([
    ('x'    , float),
    ('gluon', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='gluon')

xdata = np.empty((8, 30), xtype)
xdata['gluon'] = np.arange(8)[:, None]
xdata[    'x'] = np.linspace(0, 1, 30)
    
M = np.random.randn(20, 8, 30) # xdata -> data transform

xinteg = np.empty((8, 2), xtype)
xinteg['gluon'] = np.arange(8)[:, None]
xinteg[    'x'] = [0, 1]

suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

#### CREATE GP OBJECT ####

gp = lgp.GP()

gp.addproc(kernel, 'h')
gp.addproctransf({'h': 1}, 'primitive of f', deriv='x')
gp.addproctransf({'h': 1}, 'f', deriv=(2, 'x'))
gp.addproctransf({'primitive of f': lambda x: x['x'], 'h': -1}, 'primitive of xf(x)')

gp.addx(xdata, 'xdata', proc='f')
gp.addtransf({'xdata': M}, 'data', axes=2)

gp.addx(xinteg, 'xinteg', proc='primitive of f')
gp.addtransf({'xinteg': suminteg}, 'suminteg', axes=2)

gp.addx(xinteg, 'xintegx', proc='primitive of xf(x)')
gp.addtransf({'xintegx': suminteg}, 'sumintegx', axes=2)

#### GENERATE FAKE DATA ####

prior = gp.predfromdata({
    'suminteg' : 1,
    'sumintegx': 1,
}, ['data', 'xdata'])
priorsample = next(gvar.raniter(prior))

datamean = priorsample['data']
dataerr = np.full_like(datamean, 1)
datamean = datamean + dataerr * np.random.randn(*dataerr.shape)
data = gvar.gvar(datamean, dataerr)

# check the integral is one with trapezoid rule
x = xdata['x']
y = priorsample['xdata']
checksum = np.sum((      y[:, 1:] +       y[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx x f_i(x) =', checksum)

#### FIT ####

pred = gp.predfromdata({
    'suminteg' :    1,
    'sumintegx':    1,
    'data'     : data,
}, ['data', 'xdata'])

# check the integral is one with trapezoid rule
x = xdata['x']
y = pred['xdata']
checksum = np.sum((      y[:, 1:] +       y[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx   f_i(x) =', checksum)
checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
print('sum_i int dx x f_i(x) =', checksum)

#### PLOT RESULTS ####

fig, axs = plt.subplots(1, 2, num='pdf2', clear=True, figsize=[9, 4.5])
axs[0].set_title('PDFs')
axs[1].set_title('Data')

for i in range(len(xdata)):
    y = pred['xdata'][i]
    m = gvar.mean(y)
    s = gvar.sdev(y)
    axs[0].fill_between(xdata[i]['x'], m - s, m + s, alpha=0.6, facecolor=f'C{i}')
    y2 = priorsample['xdata'][i]
    axs[0].plot(xdata[i]['x'], y2, color=f'C{i}')

m = gvar.mean(pred['data'])
s = gvar.sdev(pred['data'])
x = np.arange(len(data))
axs[1].fill_between(x, m - s, m + s, step='mid', color='lightgray')
axs[1].errorbar(x, datamean, dataerr, color='black', linestyle='', capsize=2)

fig.tight_layout()
fig.show()



"""

Check the hypothesis that the space of valid autoregressive characteristic
polynomials is a simplex with vertices (1-x)^n (1+x)^(p-n) for n = 0, 1, ..., p.

The hypothesis turned out false.

"""

import numpy as np
from matplotlib import pyplot as plt
import jax
from jax import numpy as jnp

import lsqfitgp as lgp

class AR(lgp.AR):
    
    @staticmethod
    def phi_vertices(p):
        assert int(p) == p and p >= 0, p
        roots = jnp.where(jnp.arange(p + 1)[:, None] <= jnp.arange(p), 1, -1)
        return jax.vmap(lambda r: -jnp.atleast_1d(jnp.poly(r))[1:])(roots)
        # TODO possibly not accurate for large p, see phi_from_roots,
        # it can be probably written more efficiently with binomial power
        # expansion, and the output should be integer type
    
    @staticmethod
    def phi_from_baricentric(a, vertices):
        a = jnp.asarray(a)
        vertices = jnp.asarray(vertices)
        assert a.ndim == 1 and vertices.ndim == 2
        assert a.size == vertices.shape[0]
        assert vertices.shape[0] == vertices.shape[1] + 1
        return jnp.sum(a[:, None] * vertices, 0) / jnp.sum(a)
        # TODO rewrite as matmul
    
    @classmethod
    def baricentric_from_phi(cls, phi, vertices):
        phi = cls._process_phi(phi)
        mat = jnp.concatenate([vertices.T, jnp.ones((1,) + vertices.shape[:1])])
        rhs = jnp.concatenate([phi, jnp.ones(1)])
        return jnp.linalg.solve(mat, rhs)
        # TODO mat is integer, is there a specialized exact solve for integer
        # matrices?

p = 2
N = 10000

gen = np.random.default_rng(202207221232)
a = np.abs(gen.standard_normal((N, p + 1)))
vertices = AR.phi_vertices(p)
phi = jax.vmap(AR.phi_from_baricentric, (0, None))(a, vertices)
roots = jax.vmap(AR.inverse_roots_from_phi)(phi)

fig, ax = plt.subplots(num='convexphi', clear=True)

r = roots.reshape(-1)
ax.plot(r.real, r.imag, '.', alpha=.1)

theta = np.linspace(0, 2 * np.pi, 1000)
ax.plot(np.cos(theta), np.sin(theta), '--k')

ax.set_title(f'p = {p}')

fig.show()



# lsqfitgp/tests/test_autoreg.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
from numpy.polynomial import polynomial
from scipy import linalg
import pytest

sys.path.insert(0, '.')
import lsqfitgp as lgp
import util

gen = np.random.default_rng(202207202302)

def gen_ar_acf(p):
    if p:
        mod = gen.uniform(1.1, 10, p)
        phase = gen.uniform(0, 2 * np.pi, p)
        root = mod * np.exp(1j * phase)
        ampl = np.abs(gen.standard_normal(p))
        tau = np.arange(p + 1)
        return np.sum(ampl * root ** -tau[:, None], axis=1).real
    else:
        return np.abs(gen.standard_normal(1))

plist = [0, 1, 2, 10, 30, 100]

def test_gen_ar_acf():
    for p in plist:
        acf = gen_ar_acf(p)
        assert acf.ndim == 1 and acf.size == 1 + p
        mat = linalg.toeplitz(acf)
        w = linalg.eigvalsh(mat)
        assert np.min(w) >= -np.max(w) * len(mat) * np.finfo(float).eps

def test_yule_walker_inv():
    for p in plist:
        acf = gen_ar_acf(p)
        phi = lgp.AR.phi_from_gamma(acf)
        acf2 = lgp.AR.gamma_from_phi(phi)
        acf /= acf[0]
        acf2 /= acf2[0]
        util.assert_allclose(acf2, acf, rtol=1e-12)

def test_yule_walker_inv_extend():
    for p in plist:
        acf = gen_ar_acf(p)
        phi = lgp.AR.phi_from_gamma(acf)
        acf2 = lgp.AR.gamma_from_phi(phi)
        phi3 = np.pad(phi, (0, 1 + p))
        acf3 = lgp.AR.gamma_from_phi(phi3)
        util.assert_allclose(acf3[:len(acf2)], acf2, rtol=1e-14)

def test_yule_walker_inv_evolve():
    for p in plist:
        acf = gen_ar_acf(p)
        phi = lgp._kernels.AR.phi_from_gamma(acf)
        phi2 = np.pad(phi, (0, 1 + p))
        acf2 = lgp.AR.gamma_from_phi(phi2)
        acf3 = lgp.AR.extend_gamma(acf2[:1 + p], phi, 1 + p)
        util.assert_allclose(acf3, acf2, atol=1e-300, rtol=1e-12)

def test_yule_walker_inv_0():
    acf = lgp._kernels.AR.gamma_from_phi(np.empty(0))
    util.assert_allclose(acf, [1], rtol=1e-15)

def test_yule_walker_inv_1():
    bound = 1 - 1e-8
    phi = gen.uniform(-bound, bound)
    acf = lgp.AR.gamma_from_phi([phi])
    acf2 = 1 / ((1 - phi) * (1 + phi)) * phi ** np.arange(2)
    util.assert_allclose(acf, acf2, rtol=1e-15)

def test_phase_degeneracy():
    phases = [
        [[1], [-1]],
        [[1], [1 + 2 * np.pi]],
        [[1], [1 - 2 * np.pi]],
        [[1, -1], [1, 1]],
        [[0.01], [0.01 + 4 * np.pi]],
        [[0.01], [0.01 - 4 * np.pi]],
    ]
    lag = np.arange(100)
    def lnc(ph):
        return 0.1 + 1j * np.array(ph)
    for ph1, ph2 in phases:
        c1 = lgp.AR(slnr=[], lnc=lnc(ph1))(0, lag)
        c2 = lgp.AR(slnr=[], lnc=lnc(ph2))(0, lag)
        util.assert_allclose(c2, c1, atol=0, rtol=1e-13)

def test_real_complex():
    lag = np.arange(100)
    for r in np.logspace(-5, 0, 10):
        for n in range(3):
            for m in range(2):
                add = list(np.arange(1, m + 1) * 0.1 + 1j)
                c1 = lgp.AR(slnr=2 * n * [r], lnc=add)(0, lag)
                c2 = lgp.AR(slnr=[], lnc=add + n * [r])(0, lag)
                util.assert_allclose(c2, c1, atol=0, rtol=1e-7)

def test_ar0():
    lag = np.arange(100)
    acf = np.where(lag, 0, 1)
    params = [
        dict(phi=[], maxlag=lag.size),
        dict(gamma=[1], maxlag=lag.size),
        dict(slnr=[], lnc=[]),
    ]
    for kw in params:
        for norm in range(1):
            c = lgp.AR(**kw, norm=norm)(0, lag)
            util.assert_allclose(c, acf, atol=0, rtol=0)

def test_ar1():
    lag = np.arange(100)
    for phi in np.logspace(-5, -0.001, 10):
        acf = 1 / ((1 - phi) * (1 + phi)) * phi ** lag
        params = [
            dict(phi=[phi], maxlag=lag.size),
            dict(gamma=acf[:2], maxlag=lag.size),
            dict(slnr=[-np.log(phi)], lnc=[]),
        ]
        for kw in params:
            for norm in range(1):
                c = lgp.AR(**kw, norm=norm)(0, lag)
                den = acf[0] if norm else 1
                util.assert_allclose(c, acf / den, atol=1e-300, rtol=1e-12)

# def test_ar2():
#     p = 2
#     vertices = np.stack([
#         lgp.AR.phi_from_roots(l * [-0.] + (p - l) * [0], [])
#         for l in range(p + 1)
#     ])
#     a = np.abs(gen.standard_normal(p + 1))
#     phi = np.sum(a[:, None] * vertices, 0) / np.sum(a)
    # TODO formula for the correlation?

def test_zero_slnr():
    for p in range(1, 10):
        for s in [1, -1]:
            p1 = lgp.AR.phi_from_roots(p * [s * 0.], [])
            p2 = -np.atleast_1d(np.poly(p * [s]))[1:]
            np.testing.assert_equal(p1, p2)

def test_ma_norm():
    x = np.arange(10)[:, None]
    w = [0.1, 0.3, -0.6]
    k1 = lgp.MA(w=w)
    k2 = lgp.MA(w=w, norm=True)
    v1 = k1(x, x.T)
    v2 = k2(x, x.T)
    v2 *= np.sum(np.square(w))
    util.assert_allclose(v1, v2, rtol=1e-15)

def test_ar_invalid_argset():
    kernel = lgp.AR(phi=[1, 2, 3])
    with pytest.raises(ValueError):
        kernel([1,2,3],[3,2,1])

def test_ar_norm():
    for kw in [
        dict(phi=[1.4, 0.3], maxlag=100),
        dict(slnr=[0.1], lnc=[0.2 - 1.j])
    ]:
        k1 = lgp.AR(**kw)
        k2 = lgp.AR(**kw, norm=True)
        x = np.arange(10)[:, None]
        v1 = k1(x, x.T)
        v2 = k2(x, x.T)
        v2 *= v1[0, 0]
        util.assert_allclose(v1, v2, rtol=1e-14)    

# TODO
# test reflection of extend_gamma



import functools

import jax
from jax import numpy as jnp

import lsqfitgp as lgp

indentation = 0
def trace(fun):
    if hasattr(fun, '_traced'):
        return fun
    @functools.wraps(fun)
    def tracefun(*args, **kw):
        global indentation
        space = '    ' * indentation
        arglist = ', '.join(list(map(repr, args)) + [f'{k}={v!r}' for k, v in kw.items()])
        print(space + fun.__name__ + '(' + arglist + ')')
        indentation += 1
        rt = fun(*args, **kw)
        indentation -= 1
        out = ('\n' + space).join(repr(rt).split('\n'))
        print(space + '-> ' + out)
        return rt
    tracefun._traced = True
    return tracefun

def f(x):
    d = lgp._linalg.EigCutFullRank(jnp.array([[x]]))
    return d.quad(jnp.array([2.]), jnp.array([3.]))

g = jax.grad(f)
# print(g(1.))

@jax.custom_jvp
def f(x):
    return jnp.sum(x)

@f.defjvp
def fjvp(primals, tangents):
    x, = primals
    xdot, = tangents
    return f(x), jnp.ones(len(x)).reshape(xdot.shape) @ ((jnp.eye(len(x)) @ xdot).T / jnp.ones(len(x)))

print(jax.jacfwd(f)(jnp.ones(3)))
print(jax.jacrev(f)(jnp.ones(3)))



# lsqfitgp/examples/n.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE N.

    Where we wonder how much a derivative is allowed to do her own
    business compared to her mistress.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

x = np.linspace(-10, 10, 200)
derivs = [0, 1, 2]

gp = lgp.GP(lgp.ExpQuad(scale=2))
for d in derivs:
    gp.addx(x, d, deriv=d)

u = gp.prior()

fig, ax = plt.subplots(num='n', clear=True)

ax.axhline(0)

colors = dict()
for deriv in derivs:
    m = gvar.mean(u[deriv])
    s = gvar.sdev(u[deriv])
    patch = ax.fill_between(x, m - s, m + s, label=f'deriv {deriv}', alpha=0.5)
    colors[deriv] = patch.get_facecolor()[0]
    
for sample in gvar.raniter(u, 1):
    for deriv in derivs:
        ax.plot(x, sample[deriv], '-', color=colors[deriv])

ax.legend(loc='best')
ax.grid(linestyle=':')

fig.show()



import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np

alpha = 1

gp = lgp.GP()
gp.addproc(lgp.Gibbs(scalefun=lambda x: x + 1e-16), 'f')
gp.addprocrescale(lambda x: x ** (alpha + 1) / (alpha + 2), 'g', 'f')
gp.addprocderiv(1, 'h', 'g')

x = np.linspace(0, 1, 100)[1:]
gp.addx(x, 'x', proc='h')
cov = gp.prior('x', raw=True)
sdev = np.sqrt(np.diag(cov))

fig, ax = plt.subplots(num='priortest', clear=True)

ax.plot(x, sdev, label='prior sdev')
ax.plot(x, x ** alpha, label=f'$x^{{{alpha}}}$')
ax.legend()

fig.show()



# lsqfitgp/tests/test_copula_beta.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""Test the copula.beta module"""

import sys

from jax import test_util
from scipy import stats
import pytest
import numpy as np

sys.path.insert(0, '.')
from lsqfitgp.copula import _beta

def test_grad():
    test_util.check_grads(lambda y: _beta.betaincinv(2.5, 1.3, y), (0.3,), 2)

@pytest.mark.xfail
def test_grad_ab():
    test_util.check_grads(_beta.betaincinv, (2.5, 1.3, 0.3), 1)

def test_ppf():
    q = 0.43
    a = 3.6
    b = 2.1
    np.testing.assert_allclose(stats.beta.ppf(q, a, b), _beta.beta.ppf(q, a, b), rtol=1e-6)



# lsqfitgp/examples/z.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE Z.

    Where we sail in an infinite dimensional space to sum two
    numbers.

"""

import numpy as np
from matplotlib import pyplot as plt
import gvar

import lsqfitgp as lgp

gp = lgp.GP()

gp.addproc(lgp.ExpQuad(scale= 1), 'short')
gp.addproc(lgp.ExpQuad(scale=10), 'long' )
gp.addproctransf({'short': 0.3, 'long': 1}, 'sum')

time = np.arange(30)
time_pred = np.linspace(-30, 60, 200)

def addcomps(key, time):
    gp.addx(time, key + 'short', proc='short')
    gp.addx(time, key + 'long' , proc='long' )
    gp.addx(time, key          , proc='sum'  )

addcomps('data', time)
addcomps('pred', time_pred)

print('generate data...')
prior = gp.prior(['data', 'datashort', 'datalong'])
data = next(gvar.raniter(prior))

print('prediction...')
pred = gp.predfromdata({'data': data['data']}, ['pred', 'predshort', 'predlong'])

print('sample posterior...')
mean = gvar.mean(pred)
sdev = gvar.sdev(pred)
samples = list(gvar.raniter(pred, 1))

print('figure...')
fig, axs = plt.subplots(3, 1, num='z', clear=True, figsize=[6, 7])

for ax, comp in zip(axs, ['', 'short', 'long']):
    key = 'pred' + comp
    
    m = mean[key]
    s = sdev[key]
    ax.fill_between(time_pred, m - s, m + s, alpha=0.3, color='b')
    
    for sample in samples:
        ax.plot(time_pred, sample[key], alpha=0.2, color='b')
    
    ax.plot(time, data['data' + comp], '.k')

fig.tight_layout()
fig.show()



# lsqfitgp/tests/test_gvar.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
import gvar
from jax import tree_util
import jax

sys.path.insert(0, '.')
from lsqfitgp import _patch_gvar
import util

def check_jacobian(nprim, shape, pnz=1):
    z, s = np.random.randn(2, nprim)
    x = gvar.gvar(z, np.abs(s))
    t = np.random.randn(*shape, nprim)
    t[np.random.binomial(1, 1 - pnz, t.shape).astype(bool)] = 0
    y = t @ x
    jac, indices = _patch_gvar.jacobian(y)
    y2 = _patch_gvar.from_jacobian(gvar.mean(y), jac, indices)
    util.assert_same_gvars(y, y2, atol=1e-16)

def test_jacobian():
    check_jacobian(10, (5,))
    check_jacobian(10, (2, 3))
    check_jacobian(0, (5,))
    check_jacobian(10, (0,))
    check_jacobian(10, ())
    check_jacobian(0, ())
    check_jacobian(10, (5,), 0)
    check_jacobian(100, (5,), 0.5)

def test_bdtree():
    x = gvar.BufferDict(a=[1, 2], b=[3, 4])
    l, t = tree_util.tree_flatten(x)
    y = tree_util.tree_unflatten(t, l)
    assert np.all(x.buf == y.buf)
    assert x.keys() == y.keys()
    for k in x:
        assert x.slice_shape(k) == y.slice_shape(k)

def test_bdtree_dtype():
    x = gvar.BufferDict(dict(a=0), dtype=bool)
    l, t = tree_util.tree_flatten(x)
    y = tree_util.tree_unflatten(t, l)
    assert x.dtype == y.dtype

def test_tracer():
    x = gvar.BufferDict(a=0., b=1.)
    @jax.jit
    def f(x):
        return tree_util.tree_map(lambda x: x, x)
    y = f(x)
    assert x == y



"""Fit of parton distributions functions (PDFs)

Like pdf5, but with uncertainties on M and M2"""

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar
import lsqfit

np.random.seed(20220416)

#### SETTINGS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

indices = dict(
    # quark, antiquark
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

pid  = flavor['f0']
name = flavor['f1']

nflav  = len(flavor)

# linear data
nx        = 30 # number of PDF points used for the transformation
ndata     = 10 # number of datapoints
rankmcov  =  9 # rank of the covariance matrix of the theory error

# quadratic data
nx2       = 30 # must be <= nx
ndata2    = 10
rankmcov2 =  9

#### MODEL ####
# for each PDF:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

kernel = lgp.ExpQuad(dim='x') * lgp.White(dim='pid')

# grid of points to which we apply the transformation
xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

# linear map PDF(xdata) -> data
Mcomps = np.random.randn(rankmcov, ndata, nflav, nx)
Mcomps /= np.sqrt(Mcomps.size / ndata)
Mparams = gvar.gvar(np.random.randn(rankmcov), np.full(rankmcov, 0.1))
M = lambda params: np.tensordot(params, Mcomps, 1)

# quadratic map PDF(xdata) -> data2
M2comps = np.random.randn(rankmcov2, ndata2, nflav, nx2, nx2)
M2comps /= 2 * np.sqrt(M2comps.size / ndata2)
M2comps = (M2comps + np.swapaxes(M2comps, -1, -2)) / 2
M2params = gvar.gvar(np.random.randn(rankmcov2), np.full(rankmcov2, 0.1))
M2 = lambda params: np.tensordot(params, M2comps, 1)

# endpoints of the integral for each PDF
xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

# matrix to subtract the endpoints
suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### GP OBJECT ####

gp = lgp.GP()

gp.addproc(kernel, 'h')
gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
gp.addproctransf({
    'primitive': lambda x: x['x'],
    'h'        : -1,
}, 'primitive of xf(x)')

gp.addx(xdata, 'xdata', proc='f')

# linear data (used for warmup fit)
gp.addtransf({'xdata': M(gvar.mean(Mparams))}, 'data', axes=2)

# total momentum rule
gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)

# quark sum rules
qdiff = np.array([1, -1])[:, None] # vector to subtract two quarks
for quark in 'ducs':
    idx = indices[quark] # [quark, antiquark] indices
    label = f'{quark}{quark}bar' # the one appearing in `constraints`
    xlabel = f'x{label}'
    gp.addx(xinteg[idx], xlabel, proc='primitive')
    gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)
    
#### NONLINEAR FUNCTION ####

def fcn(params):
    
    xdata = params['xdata']
    Mparams = params['Mparams']
    M2params = params['M2params']
    
    data = np.tensordot(M(Mparams), xdata, 2)
    
    # data2 = np.einsum('dfxy,fx,fy->d', M2, xdata, xdata)
    # np.einsum does not work with gvar
    xdata2 = xdata[:, None, :nx2] * xdata[:, :nx2, None]
    data2 = np.tensordot(M2(M2params), xdata2, 3)
    
    return dict(data=data, data2=data2)

prior = gp.predfromdata(constraints, ['xdata'])
prior['Mparams'] = Mparams
prior['M2params'] = M2params

#### FAKE DATA ####

trueparams = gvar.sample(prior)
truedata = gvar.BufferDict(fcn(trueparams))

dataerr = np.full_like(truedata.buf, 0.1)
datamean = truedata.buf + dataerr * np.random.randn(*dataerr.shape)
data = gvar.BufferDict(truedata, buf=gvar.gvar(datamean, dataerr))

# check sum rules approximately with trapezoid rule
def check_integrals(x, y):
    checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
    print('sum_i int dx x f_i(x) =', checksum)
    for q in 'ducs':
        idx = indices[q]
        qx = x[idx]
        qy = y[idx]
        checksum = np.sum(qdiff * (qy[:, 1:] + qy[:, :-1]) / 2 * np.diff(qx, axis=1))
        print(f'sum_i={q}{q}bar int dx f_i(x) =', checksum)

print('check integrals in fake data:')
check_integrals(xdata['x'], trueparams['xdata'])

#### FIT ####

# find the minimization starting point with a simplified version of the fit
easyfit = gp.predfromdata(dict(data=data['data'], **constraints), ['xdata'])
p0 = gvar.mean(easyfit)

fit = lsqfit.nonlinear_fit(data, fcn, prior, p0=p0, verbose=2)
print(fit.format(maxline=True, pstyle='v'))
print(fit.format(maxline=-1))

pred = fcn(fit.p)

print('check integrals in fit:')
check_integrals(xdata['x'], fit.p['xdata'])

#### PLOT RESULTS ####

fig, axs = plt.subplots(2, 3, num='pdf6', clear=True, figsize=[13, 8])
axs[0, 0].set_title('PDFs')
axs[1, 0].set_title('PDFs')
axs[0, 1].set_title('Data')
axs[1, 1].set_title('Data (quadratic)')
axs[0, 2].set_title('M parameters')
axs[1, 2].set_title('M2 parameters')

ax = axs[0, 0]

for i in range(nflav):
    
    if i >= 4:
        ax = axs[1, 0]
    
    x = xdata[i]['x']
    ypdf = fit.p['xdata'][i]
    ydata = trueparams['xdata'][i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    
    color = 'C' + str(i // 2)
    if i % 2:
        kw = dict(hatch='//////', edgecolor=color, facecolor='none')
        kwp = dict(color=color, linestyle='--')
    else:
        kw = dict(alpha=0.6, facecolor=color)
        kwp = dict(color=color)
    
    ax.fill_between(x, m - s, m + s, label=name[i], **kw)
    ax.plot(x, ydata, **kwp)

for ax in axs[:, 0]:
    ax.legend(fontsize='small')

for ax, label in zip(axs[:, 1], ['data', 'data2']):
    d = pred[label]
    m = gvar.mean(d)
    s = gvar.sdev(d)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    d = data[label]
    ax.errorbar(x, gvar.mean(d), gvar.sdev(d), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, truedata[label], drawstyle='steps-mid', color='black', label='truth')

for ax, label in zip(axs[:, 2], ['Mparams', 'M2params']):
    p = fit.p[label]
    m = gvar.mean(p)
    s = gvar.sdev(p)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    p = prior[label]
    ax.errorbar(x, gvar.mean(p), gvar.sdev(p), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, trueparams[label], drawstyle='steps-mid', color='black', label='truth')

for ax in axs[:, 1:].flat:
    ax.legend()

fig.tight_layout()
fig.show()



"""Fit of parton distributions functions (PDFs)

Like pdf7, but with more realistic PDFs"""

import time

import lsqfitgp as lgp
import numpy as np
from scipy import linalg, interpolate
from matplotlib import pyplot as plt
import gvar
import lsqfit

np.random.seed(20220416)

#### DEFINITIONS ####

qtopm = np.array([
    #  d, db,  u, ub,  s, sb,  c, cb
    [  1,  1,  0,  0,  0,  0,  0,  0], # d+ = d + dbar
    [  1, -1,  0,  0,  0,  0,  0,  0], # d- = d - dbar
    [  0,  0,  1,  1,  0,  0,  0,  0], # u+ = etc.
    [  0,  0,  1, -1,  0,  0,  0,  0], # u-
    [  0,  0,  0,  0,  1,  1,  0,  0], # s+
    [  0,  0,  0,  0,  1, -1,  0,  0], # s-
    [  0,  0,  0,  0,  0,  0,  1,  1], # c+
    [  0,  0,  0,  0,  0,  0,  1, -1], # c-
])

pmtoev = np.array([
    # d+, d-, u+, u-, s+, s-, c+, c-
    [  1,  0,  1,  0,  1,  0,  1,  0], # Sigma = sum_q q+
    [  0,  1,  0,  1,  0,  1,  0,  1], # V     = sum_q q-
    [  0, -1,  0,  1,  0,  0,  0,  0], # V3    = u- - d-
    [  0,  1,  0,  1,  0, -2,  0,  0], # V8    = u- + d- - 2s-
    [  0,  1,  0,  1,  0,  1,  0, -3], # V15   = u- + d- + s- - 3c-
    [ -1,  0,  1,  0,  0,  0,  0,  0], # T3    = u+ - d+
    [  1,  0,  1,  0, -2,  0,  0,  0], # T8    = u+ + d+ - 2s+
    [  1,  0,  1,  0,  1,  0, -3,  0], # T15   = u+ + d+ + s+ - 3c+
])

qnames  = ['d' , 'dbar', 'u' , 'ubar', 's' , 'sbar', 'c' , 'cbar']
pmnames = ['d+', 'd-'  , 'u+', 'u-'  , 's+', 's-'  , 'c+', 'c-'  ]
evnames = ['Sigma', 'V', 'V3', 'V8', 'V15', 'T3', 'T8', 'T15']
pnames  = evnames + ['g']
tpnames = ['xSigma'] + evnames[1:] + ['xg']

nflav = len(pnames)

# linear data
ndata     = 10 # number of datapoints
rankmcov  =  9 # rank of the covariance matrix of the theory error

# quadratic data
ndata2    = 10
rankmcov2 =  9

# grid used for DGLAP evolution
grid = np.array([
    1.9999999999999954e-07, # start logspace
    3.034304765867952e-07,
    4.6035014748963906e-07,
    6.984208530700364e-07,
    1.0596094959101024e-06,
    1.607585498470808e-06,
    2.438943292891682e-06,
    3.7002272069854957e-06,
    5.613757716930151e-06,
    8.516806677573355e-06,
    1.292101569074731e-05,
    1.9602505002391748e-05,
    2.97384953722449e-05,
    4.511438394964044e-05,
    6.843744918967896e-05,
    0.00010381172986576898,
    0.00015745605600841445,
    0.00023878782918561914,
    0.00036205449638139736,
    0.0005487795323670796,
    0.0008314068836488144,
    0.0012586797144272762,
    0.0019034634022867384,
    0.0028738675812817515,
    0.004328500638820811,
    0.006496206194633799,
    0.009699159574043398,
    0.014375068581090129,
    0.02108918668378717,
    0.030521584007828916,
    0.04341491741702269,
    0.060480028754447364,
    0.08228122126204893,
    0.10914375746330703, # end logspace, start linspace
    0.14112080644440345,
    0.17802566042569432,
    0.2195041265003886,
    0.2651137041582823,
    0.31438740076927585,
    0.3668753186482242,
    0.4221667753589648,
    0.4798989029610255,
    0.5397572337880445,
    0.601472197967335,
    0.6648139482473823,
    0.7295868442414312,
    0.7956242522922756,
    0.8627839323906108,
    0.9309440808717544,
    1, # end linspace
])

# grid used for data
datagrid = grid[:-1] # exclude 1 since f(1) = 0 and zero errors upset the fit
nx = len(datagrid)

# grid used for plot
gridinterp = interpolate.interp1d(np.linspace(0, 1, len(grid)), grid)
plotgrid = gridinterp(np.linspace(0, 1, 200))

#### GAUSSIAN PROCESS ####
# Ti ~ GP
#
# fi ~ GP with sdev ~ x to compensate the scale ~ 1/x
# Vi = fi'
# f(1) - f(0) = 3
# f3(1) - f3(0) = 1
# f8(1) - f8(0) = 3
# f15(1) - f15(0) = 3
#
# f1 ~ GP   (WITHOUT scale compensation)
# tf1(x) = x^(a+1)/(a+2) f1(x)   <--- scale comp. is x^(a+1) instead of x^a,
#                                     to avoid doing x^a with a < 0 in x = 0
# Sigma(x) = tf1'(x) / x    (such that x Sigma(x) ~ x^a)
# the same with f2, tf2, g
# tf12 = tf1 + tf2
# tf12(0) - tf12(1) = 1
#
# [Sigma, g, V*, T*](1) = 0

# matrix to stack processes
stackplotgrid = np.einsum('ab,ij->abij', np.eye(nflav), np.eye(len(plotgrid)))
stackdatagrid = np.einsum('ab,ij->abij', np.eye(nflav), np.eye(len(datagrid)))

# transformation from evolution to flavor basis
evtoq = linalg.inv(pmtoev @ qtopm)

hyperprior = {
    'log(scale)' : np.log(gvar.gvar(0.5, 0.5)),
    'U(alpha_Sigma)': gvar.BufferDict.uniform('U', -0.5, 0.5),
    'U(alpha_g)'    : gvar.BufferDict.uniform('U', -0.5, 0.5),
}

def makegp(hp, quick=False):
    gp = lgp.GP(checkpos=False)
    
    eps = grid[0]
    scalefun = lambda x: hp['scale'] * (x + eps)
    kernel = lgp.Gibbs(scalefun=scalefun)
    kernel_prim = kernel.rescale(scalefun, scalefun)
    
    # define Ts and Vs
    for suffix in ['', '3', '8', '15']:
        if suffix != '':
            gp.addproc(kernel, 'T' + suffix)
        gp.addproc(kernel_prim, 'f' + suffix)
        gp.addproctransf({'f' + suffix: 1}, 'V' + suffix, deriv=1)
    
    # define xSigma
    gp.addproc(kernel, 'f1')
    a = hp['alpha_Sigma']
    gp.addproctransf({'f1': lambda x: x ** (a + 1) / (a + 2)}, 'tf1')
    gp.addproctransf({'tf1': 1}, "xSigma", deriv=1)
    
    # define xg
    gp.addproc(kernel, 'f2')
    b = hp['alpha_g']
    gp.addproctransf({'f2': lambda x: x ** (b + 1) / (b + 2)}, 'tf2')
    gp.addproctransf({'tf2': 1}, "xg", deriv=1)
    
    # define primitive of xSigma + xg
    gp.addproctransf({'tf1': 1, 'tf2': 1}, 'tf12')
    
    # define a matrix of PDF values over the x grid
    for proc in tpnames:
        gp.addx(datagrid, proc + '-datagrid', proc=proc)
    gp.addtransf({
        proc + '-datagrid': stackdatagrid[i]
        for i, proc in enumerate(tpnames)
    }, 'datagrid', 1)

    # definite integrals
    for proc in ['tf12', 'f', 'f3', 'f8', 'f15']:
        gp.addx([0, 1], proc + '-endpoints', proc=proc)
        gp.addtransf({proc + '-endpoints': [-1, 1]}, proc + '-diff')
    
    # right endpoint
    for proc in tpnames:
        gp.addx(1, f'{proc}(1)', proc=proc)
    
    if not quick:
        
        # linear data (used for warmup fit)
        global M_mean
        gp.addtransf({'datagrid': M_mean}, 'data', axes=2)
    
        # define flavor basis PDFs
        gp.addproctransf({"xSigma": lambda x: 1 / x}, 'Sigma')
        gp.addproctransf({"xg": lambda x: 1 / x}, 'g')
        for qi, qproc in enumerate(qnames):
            gp.addproctransf({
                eproc: evtoq[qi, ei]
                for ei, eproc in enumerate(evnames)
            }, qproc)
    
        # define a matrix of PDF values over the plot grid
        for proc in tpnames:
            gp.addx(plotgrid, proc + '-plotgrid', proc=proc)
        gp.addtransf({
            proc + '-plotgrid': stackplotgrid[i]
            for i, proc in enumerate(tpnames)
        }, 'plotgrid', 1)

    return gp

constraints = {
    'tf12-diff': 1,
    'f-diff'   : 3,
    'f3-diff'  : 1,
    'f8-diff'  : 3,
    'f15-diff' : 3,
    'xSigma(1)': 0,
    'V(1)'     : 0,
    'V3(1)'    : 0,
    'V8(1)'    : 0,
    'V15(1)'   : 0,
    'T3(1)'    : 0,
    'T8(1)'    : 0,
    'T15(1)'   : 0,
    'xg(1)'    : 0,
}
    
#### NONLINEAR FUNCTION ####

# linear map PDF(grid) -> data
Mcomps = np.random.randn(rankmcov, ndata, nflav, nx)
Mcomps /= np.sqrt(Mcomps.size / ndata)
Mparams = gvar.gvar(np.random.randn(rankmcov), np.full(rankmcov, 0.1))
M = lambda params: np.tensordot(params, Mcomps, 1)

M_mean = M(gvar.mean(Mparams)) # used in makegp()

# quadratic map PDF(grid) -> data2
M2comps = np.random.randn(rankmcov2, ndata2, nflav, nx, nx)
M2comps /= 2 * np.sqrt(M2comps.size / ndata2)
M2comps = (M2comps + np.swapaxes(M2comps, -1, -2)) / 2
M2params = gvar.gvar(np.random.randn(rankmcov2), np.full(rankmcov2, 0.1))
M2 = lambda params: np.tensordot(params, M2comps, 1)

def fcn(params):
    datagrid = params['datagrid']
    Mparams = params['Mparams']
    M2params = params['M2params']
    
    data = np.tensordot(M(Mparams), datagrid, 2)
    
    # data2 = np.einsum('dfxy,fx,fy->d', M2, datagrid, datagrid)
    # np.einsum does not work with gvar
    grid2 = datagrid[:, None, :] * datagrid[:, :, None]
    data2 = np.tensordot(M2(M2params), grid2, 3)
    
    return dict(data=data, data2=data2)

def makeprior(gp, plot=False):
    out = ['datagrid']
    if plot:
        out += ['plotgrid']
    prior = gp.predfromdata(constraints, out)
    prior['Mparams'] = Mparams
    prior['M2params'] = M2params
    return prior

#### FAKE DATA ####

truehp = gvar.sample(hyperprior)
truegp = makegp(truehp)
trueprior = makeprior(truegp, plot=True)
trueparams = gvar.sample(trueprior)
truedata = fcn(trueparams)

dataerr = {
    k: np.full_like(v, 0.1 * (np.max(v) - np.min(v)))
    for k, v in truedata.items()
}
data = gvar.make_fake_data(gvar.gvar(truedata, dataerr))

def check_constraints(y):
    # integrate approximately with trapezoid rule
    integ = np.sum((y[:, 1:] + y[:, :-1]) / 2 * np.diff(plotgrid), 1)
    print('int dx x (Sigma(x) + g(x)) =', integ[0] + integ[-1])
    for i in range(1, 5):
        print(f'int dx {tpnames[i]}(x) =', integ[i])
    for i, name in enumerate(tpnames):
        print(f'{name}(1) =', y[i, -1])

print('\ncheck constraints in fake data:')
check_constraints(trueparams['plotgrid'])

#### FIT ####

# find the minimization starting point with a simplified version of the fit
meangp = makegp(gvar.mean(hyperprior))
easyfit = meangp.predfromdata(dict(data=data['data'], **constraints), ['datagrid'])
p0 = gvar.mean(easyfit)

hyperprior = gvar.BufferDict(hyperprior)
hypermean = gvar.mean(hyperprior.buf)
hypercov = gvar.evalcov(hyperprior.buf)
hyperchol = linalg.cholesky(hypercov)

i = 0
lasttime = time.time()
def analyzer(hp):
    global i, lasttime
    i += 1
    now = time.time()
    interval = now - lasttime
    print(f'iteration {i:3d} ({interval:#.2g} s): {hp}')
    lasttime = now

def fitargs(hp):
    analyzer(hp)
    gp = makegp(hp, quick=True)
    prior = makeprior(gp)
    args = dict(
        data = data,
        fcn = fcn,
        prior = prior,
        p0 = p0,
    )
    residuals = linalg.solve_triangular(hyperchol, hp.buf - hypermean)
    plausibility = -1/2 * (residuals @ residuals)
    return args, plausibility
    
print('\ntrue hyperparameters:')
print(truehp)

print('\nfit:')
z0 = gvar.sample(hyperprior)
fit, fithp = lsqfit.empbayes_fit(z0, fitargs)
print(fit.format(maxline=True, pstyle='v'))
print(fit.format(maxline=-1))

gp = makegp(fithp)
prior = makeprior(gp)
pred = fcn(fit.p)
fitgrid = gp.predfromfit(dict(datagrid=fit.p['datagrid'], **constraints), 'plotgrid')

print('\ncheck constraints in fit:')
check_constraints(fitgrid)

def allkeys(d):
    for k in d:
        yield k
        m = d.extension_pattern.match(k)
        if m and d.has_distribution(m.group(1)):
            yield m.group(2)

print('\nhyperparameters (true, fitted):')
for k in allkeys(fithp):
    print(f'{k}: {truehp[k]:#.2g}\t{fithp[k]:#.2g}')

#### PLOT RESULTS ####

fig, axs = plt.subplots(2, 3, num='pdf8', clear=True, figsize=[10.5, 7], gridspec_kw=dict(width_ratios=[2, 1, 1]))
axs[0, 0].set_title('PDFs')
axs[1, 0].set_title('PDFs')
axs[0, 1].set_title('Data')
axs[1, 1].set_title('Data (quadratic)')
axs[0, 2].set_title('M parameters')
axs[1, 2].set_title('M2 parameters')

for i in range(nflav):
    
    if i < 5:
        ax = axs[0, 0]
    else:
        ax = axs[1, 0]
    
    ypdf = fitgrid[i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    ax.fill_between(plotgrid, m - s, m + s, label=tpnames[i], alpha=0.6, facecolor=f'C{i}')

    ax.plot(plotgrid, trueparams['plotgrid'][i], color=f'C{i}')

    # ax.set_xscale('log')

for ax in axs[:, 0]:
    ax.legend(fontsize='small')

for ax, label in zip(axs[:, 1], ['data', 'data2']):
    d = pred[label]
    m = gvar.mean(d)
    s = gvar.sdev(d)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    d = data[label]
    ax.errorbar(x, gvar.mean(d), gvar.sdev(d), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, truedata[label], drawstyle='steps-mid', color='black', label='truth')

for ax, label in zip(axs[:, 2], ['Mparams', 'M2params']):
    p = fit.p[label]
    m = gvar.mean(p)
    s = gvar.sdev(p)
    x = np.arange(len(m))
    ax.fill_between(x, m - s, m + s, step='mid', color='lightgray', label='fit')
    p = prior[label]
    ax.errorbar(x, gvar.mean(p), gvar.sdev(p), color='black', linestyle='', capsize=2, label='data')
    ax.plot(x, trueparams[label], drawstyle='steps-mid', color='black', label='truth')

for ax in axs[:, 1:].flat:
    ax.legend()

fig.tight_layout()
fig.show()



# lsqfitgp/copula/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" Gaussian copulas for gvar """

import abc
import functools

import gvar
from jax.scipy import stats as jstats, special as jspecial
import numpy
import jax
from jax import numpy as jnp

from .. import _patch_gvar
from . import _beta, _invgamma

class CopulaFactory(metaclass=abc.ABCMeta):
    """
    Abstract base class for copula factories.

    Class to represent transformations in gvar BufferDicts, i.e., Gaussian
    copulas.

    Parameters
    ----------
    name : str
        The unique name assigned to the transformation using
        :meth:`gvar.BufferDict.add_distribution`.
    *params : scalars
        The parameters of the distribution.

    Examples
    --------
    >>> copula = gvar.BufferDict({
    ...     'transf(x)': lgp.copula.beta('transf', 1, 1),
    ...     'transf_2(y)': lgp.copula.beta('transf_2', 3, 5),
    ... })
    >>> copula['x']
    0.50(40)
    >>> copula['y']
    0.36(18)
    """
    
    def __init_subclass__(cls, **_):
        cls.params = {}
    
    @staticmethod
    @abc.abstractmethod
    def invfcn(x, *params):
        """
        :math:`F^{-1}(\\Phi(x))`, i.e., maps a Normal variable to a variable
        with the desired marginal distribution. Jax-traceable.
        """
        pass

    def __new__(cls, name, *params):
        if gvar.BufferDict.has_distribution(name):
            invfcn = gvar.BufferDict.invfcn[name]
            if getattr(invfcn, '_CopulaFactory_subclass_', None) is not cls:
                raise ValueError(f'distribution {name} already defined')
            if cls.params[name] != params:
                raise ValueError(f'Attempt to overwrite existing {cls.__name__} distribution with name {name}')
                # cls.params is not updated by
                # gvar.BufferDict.del_distribution, but it is not a problem
        else:
            invfcn = lambda x: cls.invfcn(x, *params)
            invfcn = _patch_gvar.add_gvar_support(invfcn)
            invfcn._CopulaFactory_subclass_ = cls
            gvar.BufferDict.add_distribution(name, invfcn)
            cls.params[name] = params
        return gvar.gvar(0, 1)
    
class beta(CopulaFactory):
    """
    https://en.wikipedia.org/wiki/Beta_distribution
    """
    
    def __new__(cls, name, alpha, beta):
        return super().__new__(cls, name, alpha, beta)
    
    @staticmethod
    def invfcn(x, alpha, beta):
        return _beta.beta.ppf(jstats.norm.cdf(x), a=alpha, b=beta)

class invgamma(CopulaFactory):
    """
    https://en.wikipedia.org/wiki/Inverse-gamma_distribution
    """
    
    def __new__(cls, name, alpha, beta):
        return super().__new__(cls, name, alpha, beta)

    @staticmethod
    def _invgamma_ppf_norm_cdf_large_negative_x_1(x):
        return 1 / (1/2 * jnp.log(2 * jnp.pi) + 1/2 * jnp.square(x) + jnp.log(-x))
        # Φ(x) ≈ -1/√2π exp(-x²/2)/x
        # Q(a, x) ≈ x^(a-1) e^-x / Γ(a)
        # invgamma.ppf(x, a) = 1 / Q⁻¹(a, x)
    
    @staticmethod
    def _invgamma_ppf_norm_cdf_large_negative_x_2(x, a):
        x0 = 1/2 * jnp.log(2 * jnp.pi) + 1/2 * jnp.square(x) + jnp.log(-x)
        x1 = x0 - (-(a - 1) * jnp.log(x0) + jspecial.gammaln(a)) / (1 - (a - 1) / x0)
        return 1 / x1
        # compared to _1, this adds one newton step for Q⁻¹(a, x)

    @classmethod
    def invfcn(cls, x, alpha, beta):
        x = jnp.asarray(x)
        x = x.astype(jnp.sin(jnp.empty(0, x.dtype)).dtype)
        boundary = 37 if x.dtype == jnp.float64 else 12
        return jnp.piecewise(x, 
            [x < -boundary, x > 0],
            [
                lambda x: beta * cls._invgamma_ppf_norm_cdf_large_negative_x_2(x, a=alpha),
                lambda x: _invgamma.invgamma.isf(jstats.norm.cdf(-x), a=alpha, scale=beta),
                lambda x: _invgamma.invgamma.ppf(jstats.norm.cdf(x), a=alpha, scale=beta),
            ],
        )



# lsqfitgp/__init__.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""
Module to fit Gaussian processes

See the manual at https://gattocrucco.github.io/lsqfitgp/docs
"""

__version__ = '0.15'

from . import _patch_jax # keep first due to pre-import jax configs
from . import _patch_gvar

from ._GP import *
from ._Kernel import *
from ._kernels import *
from ._array import *
from ._fit import *
from ._Deriv import *
from ._fastraniter import *
from . import copula



"""Fit of parton distribution functions (PDFs)

Like pdf3, but with hyperparameters"""

import warnings

import lsqfitgp as lgp
import numpy as np
from matplotlib import pyplot as plt
import gvar

np.random.seed(20220416)

warnings.filterwarnings('ignore', r'total derivative orders')

#### COMPONENTS ####

flavor = np.array([
    ( 1, 'd'    ), # 0
    (-1, 'dbar' ), # 1
    ( 2, 'u'    ), # 2
    (-2, 'ubar' ), # 3
    ( 3, 's'    ), # 4
    (-3, 'sbar' ), # 5
    ( 4, 'c'    ), # 6
    (-4, 'cbar' ), # 7
    (21, 'gluon'), # 8
], 'i8, U16')

pid  = flavor['f0']
name = flavor['f1']

nflav = len(flavor)
nx    = 30
ndata = 20

indices = dict(
    d = [0, 1],
    u = [2, 3],
    s = [4, 5],
    c = [6, 7],
)

#### MODEL ####
# for each component of the proton:
# h ~ GP
# f = h''  (f is the PDF)
# for the momentum sum rule:
# int_0^1 dx x f(x) = [xh'(x) - h(x)]_0^1
# for the flavor sum rules:
# int_0^1 dx (f_i(x) - f_j(x)) = [h_i'(x) - h_j'(x)]_0^1

xtype = np.dtype([
    ('x'  , float),
    ('pid', int  ),
])

# fitting both scale and variance with weak priors is a bad idea, but whatever
hyperprior = {
    'log(scale)': gvar.log(gvar.gvar(0.5, 0.5)),
    'log(ampl)' : gvar.log(gvar.gvar(1, 1)),
    'log(beta)': gvar.log(gvar.gvar(4, 2)),
}
def makekernel(hp):
    return hp['ampl'] ** 2 * lgp.Cauchy(dim='x', scale=hp['scale'], beta=hp['beta']) * lgp.White(dim='pid')

xdata = np.empty((nflav, nx), xtype)
xdata['pid'] = pid[:, None]
xdata[  'x'] = np.linspace(0, 1, nx)

M = np.random.randn(ndata, nflav, nx) # transformation PDF(X) -> data

xinteg = np.empty((nflav, 2), xtype)
xinteg['pid'] = pid[:, None]
xinteg[  'x'] = [0, 1]

suminteg = np.empty(xinteg.shape)
suminteg[:, 0] = -1
suminteg[:, 1] =  1

qdiff = np.array([1, -1])[:, None]

constraints = {
    'momrule': 1,
    'uubar'  : 2,
    'ddbar'  : 1,
    'ccbar'  : 0,
    'ssbar'  : 0,
}

#### CREATE GP OBJECT ####

def makegp(hp):
    
    gp = lgp.GP()

    kernel = makekernel(hp)
    gp.addproc(kernel, 'h')
    gp.addproctransf({'h': 1}, 'primitive', deriv='x'     )
    gp.addproctransf({'h': 1}, 'f'        , deriv=(2, 'x'))
    gp.addproctransf({
        'primitive': lambda x: x['x'],
        'h'        : -1,
    }, 'primitive of xf(x)')

    # data
    gp.addx(xdata, 'xdata', proc='f')
    gp.addtransf({'xdata': M}, 'data', axes=2)

    # total momentum rule
    gp.addx(xinteg, 'xmomrule', proc='primitive of xf(x)')
    gp.addtransf({'xmomrule': suminteg}, 'momrule', axes=2)
    
    # quark sum rules
    for quark in 'ducs':
        idx = indices[quark]
        label = f'{quark}{quark}bar' # the one appearing in `constraints`
        xlabel = f'x{label}'
        gp.addx(xinteg[idx], xlabel, proc='primitive')
        gp.addtransf({xlabel: suminteg[idx] * qdiff}, label, axes=2)

    return gp

#### GENERATE FAKE DATA ####

hpsample = next(gvar.raniter(hyperprior))
gp = makegp(hpsample)
prior = gp.predfromdata(constraints, ['data', 'xdata'])
priorsample = next(gvar.raniter(prior))

datamean = priorsample['data']
dataerr = np.full_like(datamean, 1)
datamean = datamean + dataerr * np.random.randn(*dataerr.shape)
data = gvar.gvar(datamean, dataerr)

# check sum rules approximately with trapezoid rule
def check_integrals(x, y):
    checksum = np.sum(((y * x)[:, 1:] + (y * x)[:, :-1]) / 2 * np.diff(x, axis=1))
    print('sum_i int dx x f_i(x) =', checksum)
    for q in 'ducs':
        idx = indices[q]
        qx = x[idx]
        qy = y[idx]
        checksum = np.sum(qdiff * (qy[:, 1:] + qy[:, :-1]) / 2 * np.diff(qx, axis=1))
        print(f'sum_i={q}{q}bar int dx f_i(x) =', checksum)

print('check integrals in fake data:')
check_integrals(xdata['x'], priorsample['xdata'])

#### FIT ####
information = dict(constraints)
information.update({
    'data': data,
})

fit = lgp.empbayes_fit(hyperprior, makegp, information, raises=False, jit=True, verbosity=3, forward=True)

print('hyperparameters:')
for k in fit.p.all_keys():
    d = fit.p[k] - hpsample[k]
    print(f'{k:10}: fit {fit.p[k]} true {hpsample[k]:5.2g} diff {d}')

gp = makegp(gvar.mean(fit.p))
pred = gp.predfromdata(information, ['data', 'xdata'])

print('check integrals in fit:')
check_integrals(xdata['x'], pred['xdata'])

#### PLOT RESULTS ####

fig, axs = plt.subplots(1, 2, num='pdf4', clear=True, figsize=[9, 4.5])
axs[0].set_title('PDFs')
axs[1].set_title('Data')

for i in range(nflav):
    
    x = xdata[i]['x']
    ypdf = pred['xdata'][i]
    ydata = priorsample['xdata'][i]
    m = gvar.mean(ypdf)
    s = gvar.sdev(ypdf)
    
    color = 'C' + str(i // 2)
    if i % 2:
        kw = dict(hatch='//////', edgecolor=color, facecolor='none')
        kwp = dict(color=color, linestyle='--')
    else:
        kw = dict(alpha=0.6, facecolor=color)
        kwp = dict(color=color)
    
    axs[0].fill_between(x, m - s, m + s, label=name[i], **kw)
    axs[0].plot(x, ydata, **kwp)

axs[0].legend(fontsize='small')

m = gvar.mean(pred['data'])
s = gvar.sdev(pred['data'])
x = np.arange(len(data))
axs[1].fill_between(x, m - s, m + s, step='mid', color='lightgray')
axs[1].errorbar(x, datamean, dataerr, color='black', linestyle='', capsize=2)

fig.tight_layout()
fig.show()



import jax
from jax import numpy as jnp
import numpy as np
from scipy import linalg

# adapted from SuperGauss/DurbinLevinson.h

def chol_solve(t, b):
    b = jnp.asarray(b)
    t = jnp.asarray(t)
    assert t.ndim == 1 and b.ndim == 2
    assert len(t) == len(b)

    ilb = jnp.zeros_like(b)
    n = len(t)
    phi1 = jnp.zeros(n)
    phi2 = jnp.zeros(n)
    
    nu = t[0]
    tlag = jnp.roll(t, -1)
    sqrt_nu = jnp.sqrt(nu)
    l0 = jnp.zeros(n).at[0].set(1 / sqrt_nu)
    
    ilb = ilb.at[0, :].set(l0 @ b)
    
    for i in range(1, n):
        
        # now:
        # phi1[i - 1:] == 0
        # phi2[i - 2:] == 0
        
        pi = i - 1
        rp = phi2 @ tlag
        phi1 = phi1.at[pi].set((tlag[pi] - rp) / nu)
        phi1 = phi1 - phi1[pi] * phi2
        
        # now:
        # phi1[i:] == 0
        # phi2[i - 1:] == 0

        nu = nu * (1 - phi1[pi]) * (1 + phi1[pi])
        
        phi2 = jnp.roll(phi1[::-1], i)

        # i-th row of L^-1
        li = -phi2.at[i].set(-1) / jnp.sqrt(nu)
        
        ilb = ilb.at[i, :].set(li @ b)

    return ilb

n = 10
m = 3
x = np.linspace(0, 3, n)
t = np.exp(-1/2 * (x - x[0]) ** 2)
b = np.random.randn(n, m)
a = linalg.toeplitz(t)
l = linalg.cholesky(a, lower=True)
ilb1 = chol_solve(t, b)
ilb2 = linalg.solve_triangular(l, b, lower=True)
np.testing.assert_allclose(ilb1, ilb2, rtol=1e-8)



# lsqfitgp/_kernels/_basic.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys
import re
import collections

import numpy
import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _special
from .. import _patch_jax
from .. import _Kernel
from .._Kernel import kernel, stationarykernel, isotropickernel

@isotropickernel(derivable=True, input='raw')
def Constant(x, y):
    """
    Constant kernel.
    
    .. math::
        k(x, y) = 1
    
    This means that all points are completely correlated, thus it is equivalent
    to fitting with a horizontal line. This can be seen also by observing that
    1 = 1 x 1.
    """
    return jnp.ones(jnp.broadcast_shapes(x.shape, y.shape))
    
@isotropickernel(derivable=False, input='raw')
def White(x, y):
    """
    White noise kernel.
    
    .. math::
        k(x, y) = \\begin{cases}
            1 & x = y     \\\\
            0 & x \\neq y
        \\end{cases}
    """
    return _Kernel.prod_recurse_dtype(lambda x, y: x == y, x, y).astype(int)
    # TODO maybe StructuredArray should support equality and other operations

@isotropickernel(derivable=True)
def ExpQuad(r2):
    """
    Exponential quadratic kernel.
    
    .. math::
        k(r) = \\exp \\left( -\\frac 12 r^2 \\right)
    
    It is smooth and has a strict typical lengthscale, i.e., oscillations are
    strongly suppressed under a certain wavelength, and correlations are
    strongly suppressed over a certain distance.
    
    Reference: Rasmussen and Williams (2006, p. 83).
    """
    return jnp.exp(-1/2 * r2)

def _dot(x, y):
    return _Kernel.sum_recurse_dtype(lambda x, y: x * y, x, y)
    
@kernel(derivable=True)
def Linear(x, y):
    """
    Dot product kernel.
    
    .. math::
        k(x, y) = x \\cdot y = \\sum_i x_i y_i
    
    In 1D it is equivalent to fitting with a line passing by the origin.

    Reference: Rasmussen and Williams (2006, p. 89).
    """
    return _dot(x, y)

@isotropickernel(derivable=lambda gamma=1: gamma == 2)
def GammaExp(r2, gamma=1):
    """
    Gamma exponential kernel.
    
    .. math::
        k(r) = \\exp(-r^\\gamma), \\quad
        \\gamma \\in (0, 2]
    
    For :math:`\\gamma = 2` it is the squared exponential kernel, for
    :math:`\\gamma = 1` (default) it is the Matérn 1/2 kernel, for
    :math:`\\gamma \\to 0` it tends to white noise plus a constant. The process
    is differentiable only for :math:`\\gamma = 2`, however as :math:`\\gamma`
    gets closer to 2 the variance of the non-derivable component goes to zero.

    Reference: Rasmussen and Williams (2006, p. 86).
    """
    with _patch_jax.skipifabstract():
        assert 0 < gamma <= 2, gamma
    nondiff = jnp.exp(-(r2 ** (gamma / 2)))
    diff = jnp.exp(-r2)
    return jnp.where(gamma == 2, diff, nondiff)
    # I need to keep separate the case where derivatives w.r.t. r2 could be
    # computed because the second derivative at x=0 of x^p with floating
    # point p is nan.
    
    # TODO extend to gamma=0, the correct limit is
    # e^-1 constant + (1 - e^-1) white noise. Use lax.switch.

    # TODO derivatives w.r.t. gamma at gamma==2 are probably broken, although
    # I guess they are not needed since it's on the boundary of the domain
    
@kernel(derivable=True)
def NNKernel(x, y, sigma0=1):
    """
    Neural network kernel.
    
    .. math::
        k(x, y) = \\frac 2 \\pi
        \\arcsin \\left( \\frac
        {
            2 (q + x \\cdot y)
        }{
            (1 + 2 (q + x \\cdot x))
            (1 + 2 (q + y \\cdot y))
        }
        \\right),
        \\quad q = \\texttt{sigma0}^2
    
    Kernel which is equivalent to a neural network with one infinite hidden
    layer with Gaussian priors on the weights and error function response. In
    other words, you can think of the process as a superposition of sigmoids
    where `sigma0` sets the dispersion of the centers of the sigmoids.
    
    Reference: Rasmussen and Williams (2006, p. 90).
    """
    
    # TODO the `2`s in the formula are a bit arbitrary. Remove them or give
    # motivation relative to the precise formulation of the neural network.
    with _patch_jax.skipifabstract():
        assert 0 < sigma0 < jnp.inf
    q = sigma0 ** 2
    denom = (1 + 2 * (q + _dot(x, x))) * (1 + 2 * (q + _dot(y, y)))
    return 2/jnp.pi * jnp.arcsin(2 * (q + _dot(x, y)) / denom)
    
    # TODO this is not fully equivalent to an arbitrary transformation on the
    # augmented vector even if x and y are transformed, unless I support q
    # being a vector or an additional parameter.

@kernel(maxdim=sys.maxsize)
def Gibbs(x, y, scalefun=lambda x: 1):
    """
    Gibbs kernel.
    
    .. math::
        k(x, y) = \\sqrt{ \\frac {2 s(x) s(y)} {s(x)^2 + s(y)^2} }
        \\exp \\left( -\\frac {(x - y)^2} {s(x)^2 + s(y)^2} \\right),
        \\quad s = \\texttt{scalefun}.
    
    Kernel which in some sense is like a Gaussian kernel where the scale
    changes at every point. The scale is computed by the parameter `scalefun`
    which must be a callable taking the x array and returning a scale for each
    point. By default `scalefun` returns 1 so it is a Gaussian kernel.
    
    Consider that the default parameter `scale` acts before `scalefun`, so
    for example if `scalefun(x) = x` then `scale` has no effect. You should
    include all rescalings in `scalefun` to avoid surprises.
    
    Reference: Rasmussen and Williams (2006, p. 93).
    """
    sx = scalefun(x)
    sy = scalefun(y)
    with _patch_jax.skipifabstract():
        assert jnp.all(sx > 0)
        assert jnp.all(sy > 0)
    denom = sx ** 2 + sy ** 2
    factor = jnp.sqrt(2 * sx * sy / denom)
    distsq = _Kernel.sum_recurse_dtype(lambda x, y: (x - y) ** 2, x, y)
    return factor * jnp.exp(-distsq / denom)

@stationarykernel(derivable=True, forcekron=True)
def Periodic(delta, outerscale=1):
    """
    Periodic Gaussian kernel.
    
    .. math::
        k(\\Delta) = \\exp \\left(
        -2 \\left(
        \\frac {\\sin(\\Delta / 2)} {\\texttt{outerscale}}
        \\right)^2
        \\right)
    
    A Gaussian kernel over a transformed periodic space. It represents a
    periodic process. The usual `scale` parameter sets the period, with the
    default `scale` = 1 giving a period of 2π, while the `outerscale` parameter
    sets the length scale of the correlations.
    
    Reference: Rasmussen and Williams (2006, p. 92).
    """
    with _patch_jax.skipifabstract():
        assert 0 < outerscale < jnp.inf
    return jnp.exp(-2 * (jnp.sin(delta / 2) / outerscale) ** 2)

@kernel(forcekron=True, derivable=False)
def Categorical(x, y, cov=None):
    """
    Categorical kernel.
    
    .. math::
        k(x, y) = \\texttt{cov}[x, y]
    
    A kernel over integers from 0 to N-1. The parameter `cov` is the covariance
    matrix of the values.
    """
        
    # TODO support sparse matrix for cov (replace jnp.asarray and numpy
    # check)
    
    assert jnp.issubdtype(x.dtype, jnp.integer)
    cov = jnp.asarray(cov)
    assert len(cov.shape) == 2
    assert cov.shape[0] == cov.shape[1]
    with _patch_jax.skipifabstract():
        assert jnp.allclose(cov, cov.T)
    return cov[x, y]

@kernel
def Rescaling(x, y, stdfun=None):
    """
    Outer product kernel.
    
    .. math::
        k(x, y) = \\texttt{stdfun}(x) \\texttt{stdfun}(y)
    
    A totally correlated kernel with arbitrary variance. Parameter `stdfun`
    must be a function that takes `x` or `y` and computes the standard
    deviation at the point. It can yield negative values; points with the same
    sign of `fun` will be totally correlated, points with different sign will
    be totally anticorrelated. Use this kernel to modulate the variance of
    other kernels. By default `stdfun` returns a constant, so it is equivalent
    to :class:`Constant`.
    
    """
    if stdfun is None:
        stdfun = lambda x: jnp.ones(x.shape)
        # do not use np.ones_like because it does not recognize StructuredArray
        # do not use x.dtype because it could be structured
    return stdfun(x) * stdfun(y)

@stationarykernel(forcekron=True, derivable=False, input='hard')
def Expon(delta):
    """
    Exponential kernel.
    
    .. math::
        k(\\Delta) = \\exp(-|\\Delta|)
    
    In 1D it is equivalent to the Matérn 1/2 kernel, however in more dimensions
    it acts separately while the Matérn kernel is isotropic.

    Reference: Rasmussen and Williams (2006, p. 85).
    """
    return jnp.exp(-delta)

    # TODO rename Laplace, write it in terms of the 1-norm directly, then do
    # TruncLaplace that reaches zero over a prescribed box. (Or truncate with
    # an option truncbox=[(l0, r0), (l1, r1), ...]).

_bow_regexp = re.compile(r'\s|[!«»"“”‘’/()\'?¡¿„‚<>,;.:-–—]')

@kernel(forcekron=True, derivable=False)
@numpy.vectorize
def BagOfWords(x, y):
    """
    Bag of words kernel.
    
    .. math::
        k(x, y) &= \\sum_{w \\in \\text{words}} c_w(x) c_w(y), \\\\
        c_w(x) &= \\text{number of times word $w$ appears in $x$}
    
    The words are defined as non-empty substrings delimited by spaces or one of
    the following punctuation characters: ! « » " “ ” ‘ ’ / ( ) ' ? ¡ ¿ „ ‚ < >
    , ; . : - – —.

    Reference: Rasmussen and Williams (2006, p. 100).
    """
    
    # TODO precompute the bags for x and y, then call a vectorized private
    # function.
    
    # TODO iterate on the shorter bag and use get on the other instead of
    # computing set intersection? Or: convert words to integers and then do
    # set intersection with sorted arrays?
    
    xbag = collections.Counter(_bow_regexp.split(x))
    ybag = collections.Counter(_bow_regexp.split(y))
    xbag[''] = 0 # why this? I can't recall
    ybag[''] = 0
    common = set(xbag) & set(ybag)
    return sum(xbag[k] * ybag[k] for k in common)

# TODO add bag of characters and maybe other text kernels

@stationarykernel(derivable=False, input='hard', forcekron=True)
def HoleEffect(delta):
    """
    
    Hole effect kernel.
    
    .. math:: k(\\Delta) = (1 - \\Delta) \\exp(-\\Delta)
    
    Reference: Dietrich and Newsam (1997, p. 1096).
    
    """
    return (1 - delta) * jnp.exp(-delta)

def _cauchy_derivable(alpha=2, **_):
    return alpha == 2

@isotropickernel(derivable=_cauchy_derivable)
def Cauchy(r2, alpha=2, beta=2):
    """
    Generalized Cauchy kernel.
    
    .. math::
        k(r) = \\left(1 + \\frac{r^\\alpha}{\\beta} \\right)^{-\\beta/\\alpha},
        \\quad \\alpha \\in (0, 2], \\beta > 0.
    
    In the geostatistics literature, the case :math:`\\alpha=2` and
    :math:`\\beta=2` (default) is known as the Cauchy kernel. In the machine
    learning literature, the case :math:`\\alpha=2` (for any :math:`\\beta`) is
    known as the rational quadratic kernel. For :math:`\\beta\\to\\infty` it is
    equivalent to ``GammaExp(gamma=alpha, scale=alpha ** (1/alpha))``, while
    for :math:`\\beta\\to 0` to ``Constant``. It is smooth only for
    :math:`\\alpha=2`.
    
    References: Gneiting and Schlather (2004, p. 273), Rasmussen and Williams
    (2006, p. 86).
    
    """
    with _patch_jax.skipifabstract():
        assert 0 < alpha <= 2, alpha
        assert 0 < beta, beta
    power = jnp.where(alpha == 2, r2, r2 ** (alpha / 2))
    # I need to keep separate the case where derivatives w.r.t. r2 could be
    # computed because the second derivative at x=0 of x^p with floating
    # point p is nan.
    return (1 + power / beta) ** (-beta / alpha)
    
    # TODO derivatives w.r.t. alpha at alpha==2 are probably broken, although
    # I guess they are not needed since it's on the boundary of the domain
    
def _causalexpquad_derivable(alpha=1):
    return alpha == 0

@isotropickernel(derivable=_causalexpquad_derivable, input='soft')
def CausalExpQuad(r, alpha=1):
    """
    Causal exponential quadratic kernel.
    
    .. math::
        k(r) = \\big(1 - \\operatorname{erf}(\\alpha r/4)\\big)
        \\exp\\left(-\\frac12 r^2 \\right)
        
    From https://github.com/wesselb/mlkernels.
    """
    with _patch_jax.skipifabstract():
        assert alpha >= 0, alpha
    return jspecial.erfc(alpha / 4 * r) * jnp.exp(-1/2 * jnp.square(r))
    # TODO taylor-expand erfc near 0 and use r2

@kernel(derivable=True, maxdim=1)
def Decaying(x, y, alpha=1):
    """
    Decaying kernel.
    
    .. math::
        k(x, y) =
        \\frac{1}{(1 + x + y)^\\alpha},
        \\quad x, y, \\alpha \\ge 0
        
    Reference: Swersky, Snoek and Adams (2014).
    """
    # TODO high dimensional version of this, see mlkernels issue #3
    with _patch_jax.skipifabstract():
        assert jnp.all(x >= 0)
        assert jnp.all(y >= 0)
    return 1 / (x + y + 1) ** alpha
    # use x + y + 1 instead of 1 + x + y because the latter is less numerically
    # accurate and symmetric for small x and y

@isotropickernel(derivable=False, input='soft')
def Log(r):
    """
    Log kernel.
    
    .. math::
        k(r) = \\log(1 + r) / r
    
    From https://github.com/wesselb/mlkernels.
    """
    return jnp.log1p(r) / r

@kernel(forcekron=True, derivable=True)
def Taylor(x, y):
    """
    Exponential-like power series kernel.
    
    .. math::
        k(x, y) = \\sum_{k=0}^\\infty \\frac {x^k}{k!} \\frac {y^k}{k!}
        = I_0(2 \\sqrt{xy})
    
    It is equivalent to fitting with a Taylor series expansion in zero with
    independent priors on the coefficients k with mean zero and standard
    deviation 1/k!.
    """
    
    # TODO reference? Maybe it's called bessel kernel in the literature?
    # => nope, bessel kernel is the J_v one
        
    # TODO what is the "natural" extension of this to multidim? Is forcekron
    # appropriate?
    
    # TODO probably the rescaled version of this (e^-x) makes more sense
    
    mul = x * y
    val = 2 * jnp.sqrt(jnp.abs(mul))
    return jnp.where(mul >= 0, jspecial.i0(val), _special.j0(val))



import functools

class cippa: pass
class lippa: pass
class fagiolino(cippa, lippa): pass

@functools.singledispatch
def f(_):
    print('not dispatched')

@f.register
def _(_: lippa):
    print('lippa')

f(None)
f(fagiolino())



from lsqfitgp import _toeplitz
import numpy as np
from scipy import linalg

def rtol(x):
    return 1e-4 if x.dtype == np.float32 else 1e-8
    
def atol(x):
    return np.finfo(x.dtype).eps * np.max(x)

def maket(n):
    x = np.linspace(0, 3, n)
    t = np.pi * np.exp(-1/2 * x ** 2)
    dx = x[1] - x[0] if len(x) > 1 else np.empty(())
    difft = t[:-1] * np.expm1(-dx * (x[:-1] + dx / 2))
    np.testing.assert_allclose(np.diff(t), difft, rtol=rtol(t), atol=atol(x))
    return t, difft

def makeb(*shape):
    assert len(shape) <= 2
    n = shape[0]
    gen = np.random.default_rng(n + 202206041955)
    return gen.standard_normal(shape)

def catch(job):
    try:
        return job()
    except np.linalg.LinAlgError as exc:
        print(', '.join(exc.args))
        
def checkchol(t, b, lb, prefix):
    if l is None:
        pass
    elif not np.all(np.isfinite(l)):
        order, = np.nonzero(np.any(np.isnan(l), 0))
        order = order[0]
        print(f'{order + 1}-th minor is not positive definite')
    else:
        assert l.dtype == m.dtype, (l.dtype, m.dtype)
        ml = l @ l.T
        np.testing.assert_allclose(ml, m, rtol=rtol(ml))
        e = np.linalg.norm(ml - m, 2)
        print(f'error {prefix}: {e:.3g}')

def checkdiff(n, jit=False, eps=0, dtype='f4'):
    dtype = np.dtype(dtype)
    t, difft = maket(n, dtype)
    m = linalg.toeplitz(t)
    assert m.dtype == dtype
    cholesky = _toeplitz.cholesky_jit if jit else _toeplitz.cholesky
    l1 = catch(lambda: cholesky(t, diageps=eps))
    checkchol(l1, m, 'w/o diff')
    l2 = catch(lambda: cholesky(t, difft=difft, diageps=eps))
    checkchol(l2, m, 'w/ diff')

def checkchol2(t, b, ub, prefix):
    if ub is None:
        pass
    elif not np.all(np.isfinite(ub)):
        ub = ub.reshape(ub.shape[0], -1)
        order, = np.nonzero(np.any(np.isnan(ub), 1))
        order = order[0]
        print(f'{order + 1}-th minor is not positive definite')
    else:
        qref = b.T @ linalg.matmul_toeplitz(t, b)
        q = ub.T @ ub
        erel = (q - qref) / qref
        print(f'relative error on quadratic form {prefix}: {erel:.2g}')

def checkdiff2(n, jit=True, eps=0, dtype='f4'):
    t, difft = maket(n)
    b = makeb(n)
    cholesky = _toeplitz.cholesky_jit if jit else _toeplitz.cholesky
    ub1 = catch(lambda: cholesky(t.astype(dtype), b.astype(dtype), diageps=eps, lower=False))
    checkchol2(t, b, ub1, 'w/o diff')
    ub2 = catch(lambda: cholesky(t.astype(dtype), b.astype(dtype), difft=difft.astype(dtype), diageps=eps, lower=False))
    checkchol2(t, b, ub2, 'w/ diff')



#! /usr/bin/env python3
#
def r8mat_print ( m, n, a, title ):

#*****************************************************************************80
#
## r8mat_print prints an R8MAT.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    31 August 2014
#
#  Author:
#
#    John Burkardt
#
#  Input:
#
#    integer M, the number of rows in A.
#
#    integer N, the number of columns in A.
#
#    real A(M,N), the matrix.
#
#    string TITLE, a title.
#
  r8mat_print_some ( m, n, a, 0, 0, m - 1, n - 1, title )

  return

def r8mat_print_test ( ):

#*****************************************************************************80
#
## r8mat_print_test tests r8mat_print.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    10 February 2015
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'r8mat_print_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  r8mat_print prints an R8MAT.' )

  m = 4
  n = 6
  v = np.array ( [ \
    [ 11.0, 12.0, 13.0, 14.0, 15.0, 16.0 ], 
    [ 21.0, 22.0, 23.0, 24.0, 25.0, 26.0 ], 
    [ 31.0, 32.0, 33.0, 34.0, 35.0, 36.0 ], 
    [ 41.0, 42.0, 43.0, 44.0, 45.0, 46.0 ] ], dtype = np.float64 )
  r8mat_print ( m, n, v, '  Here is an R8MAT:' )
#
#  Terminate.
#
  print ( '' )
  print ( 'r8mat_print_test:' )
  print ( '  Normal end of execution.' )
  return

def r8mat_print_some ( m, n, a, ilo, jlo, ihi, jhi, title ):

#*****************************************************************************80
#
## r8mat_print_some prints out a portion of an R8MAT.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    10 February 2015
#
#  Author:
#
#    John Burkardt
#
#  Input:
#
#    integer M, N, the number of rows and columns of the matrix.
#
#    real A(M,N), an M by N matrix to be printed.
#
#    integer ILO, JLO, the first row and column to print.
#
#    integer IHI, JHI, the last row and column to print.
#
#    string TITLE, a title.
#
  incx = 5

  print ( '' )
  print ( title )

  if ( m <= 0 or n <= 0 ):
    print ( '' )
    print ( '  (None)' )
    return

  for j2lo in range ( max ( jlo, 0 ), min ( jhi + 1, n ), incx ):

    j2hi = j2lo + incx - 1
    j2hi = min ( j2hi, n )
    j2hi = min ( j2hi, jhi )
    
    print ( '' )
    print ( '  Col: ' ),

    for j in range ( j2lo, j2hi + 1 ):
      print ( '%7d       ' % ( j ) ),

    print ( '' )
    print ( '  Row' )

    i2lo = max ( ilo, 0 )
    i2hi = min ( ihi, m )

    for i in range ( i2lo, i2hi + 1 ):

      print ( '%7d :' % ( i ) ),
      
      for j in range ( j2lo, j2hi + 1 ):
        print ( '%12g  ' % ( a[i,j] ) ),

      print ( '' )

  return

def r8mat_print_some_test ( ):

#*****************************************************************************80
#
## r8mat_print_some_test tests r8mat_print_some.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    31 October 2014
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'r8mat_print_some_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  r8mat_print_some prints some of an R8MAT.' )

  m = 4
  n = 6
  v = np.array ( [ \
    [ 11.0, 12.0, 13.0, 14.0, 15.0, 16.0 ], 
    [ 21.0, 22.0, 23.0, 24.0, 25.0, 26.0 ], 
    [ 31.0, 32.0, 33.0, 34.0, 35.0, 36.0 ], 
    [ 41.0, 42.0, 43.0, 44.0, 45.0, 46.0 ] ], dtype = np.float64 )
  r8mat_print_some ( m, n, v, 0, 3, 2, 5, '  Here is an R8MAT:' )
#
#  Terminate.
#
  print ( '' )
  print ( 'r8mat_print_some_test:' )
  print ( '  Normal end of execution.' )
  return

def r8vec_print ( n, a, title ):

#*****************************************************************************80
#
## r8vec_print prints an R8VEC.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    31 August 2014
#
#  Author:
#
#    John Burkardt
#
#  Input:
#
#    integer N, the dimension of the vector.
#
#    real A(N), the vector to be printed.
#
#    string TITLE, a title.
#
  print ( '' )
  print ( title )
  print ( '' )
  for i in range ( 0, n ):
    print ( '%6d:  %12g' % ( i, a[i] ) )

def r8vec_print_test ( ):

#*****************************************************************************80
#
## r8vec_print_test tests r8vec_print.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    29 October 2014
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'r8vec_print_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  r8vec_print prints an R8VEC.' )

  n = 4
  v = np.array ( [ 123.456, 0.000005, -1.0E+06, 3.14159265 ], dtype = np.float64 )
  r8vec_print ( n, v, '  Here is an R8VEC:' )
#
#  Terminate.
#
  print ( '' )
  print ( 'r8vec_print_test:' )
  print ( '  Normal end of execution.' )
  return

def t_cholesky_lower ( n, t ):

#*****************************************************************************80
#
## t_cholesky_lower: lower Cholesky factor of a compressed Toeplitz matrix.
#
#  Discussion:
#
#    The first row of the Toeplitz matrix A is supplied in T.
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = L * L'.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real T(N), the first row of the Toeplitz matrix.
#
#  Output:
#
#    real L(N,N), the lower Cholesky factor.
#
  import numpy as np

  g = np.zeros ( [ 2, n ] )

  for j in range ( 0, n ):
    g[0,j] = t[j]
  for j in range ( 1, n ):
    g[1,j] = t[j]

  l = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    l[j,0] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      l[j,i] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return l

def t_cholesky_lower_test ( ):

#*****************************************************************************80
#
## t_cholesky_lower_test tests t_cholesky_lower.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 't_cholesky_lower_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  t_cholesky_lower computes the lower Cholesky' )
  print ( '  factor of a positive semi-definite Toeplitz matrix.' )
  print ( '  The Toeplitz matrix is defined by its first row.' )

  n = 3
  t = np.array ( [ 1.0,   0.5, -0.375 ] )

  r8vec_print ( n, t, '  First row of Toeplitz matrix T:' )

  l = t_cholesky_lower ( n, t )
  r8mat_print ( n, n, l, '  Computed lower Cholesky factor L:' )

  b = np.dot ( l, l.transpose ( ) )
  r8mat_print ( n, n, b, '  Product L*L\':' )

  return

def t_cholesky_upper ( n, t ):

#*****************************************************************************80
#
## t_cholesky_upper: upper Cholesky factor of a compressed Toeplitz matrix.
#
#  Discussion:
#
#    The Toeplitz matrix A is supplied by its first row.
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = R' * R.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real T(N), the compressed Toeplitz matrix.
#
#  Output:
#
#    real R(N,N), the upper Cholesky factor.
#
  import numpy as np

  g = np.zeros ( [ 2, n ] )

  for j in range ( 0, n ):
    g[0,j] = t[j]
  for j in range ( 1, n ):
    g[1,j] = t[j]

  r = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    r[0,j] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      r[i,j] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return r

def t_cholesky_upper_test ( ):

#*****************************************************************************80
#
## t_cholesky_upper_test tests t_cholesky_upper.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 't_cholesky_upper_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  t_cholesky_upper computes the upper Cholesky' )
  print ( '  factor of a positive semi-definite Toeplitz matrix.' )
  print ( '  The Toeplitz matrix is defined by its first row.' )

  n = 3
  t = np.array ( [ 1.0,   0.5, -0.375 ] )

  r8vec_print ( n, t, '  First row of Toeplitz matrix T:' )

  r = t_cholesky_upper ( n, t )
  r8mat_print ( n, n, r, '  Computed upper Cholesky factor R:' )

  b = np.dot ( r.transpose ( ), r )
  r8mat_print ( n, n, b, '  Product R\'R:' )

  return

def timestamp ( ):

#*****************************************************************************80
#
## timestamp() prints the date as a timestamp.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license. 
#
#  Modified:
#
#    06 April 2013
#
#  Author:
#
#    John Burkardt
#
  import time

  t = time.time ( )
  print ( time.ctime ( t ) )

  return None

def toep_cholesky_lower ( n, g ):

#*****************************************************************************80
#
## toep_cholesky_lower: lower Cholesky factor of a compressed Toeplitz matrix.
#
#  Discussion:
#
#    The Toeplitz matrix A is supplied in a compressed form G.
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = L * L'.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real G(2,N), the compressed Toeplitz matrix.
#    G(1,1:N) contains the first row.
#    G(2,2:N) contains the first column.
#
#  Output:
#
#    real L(N,N), the lower Cholesky factor.
#
  import numpy as np

  l = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    l[j,0] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      l[j,i] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return l

def toep_cholesky_lower_test ( ):

#*****************************************************************************80
#
## toep_cholesky_lower_test tests toep_cholesky_lower.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    27 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'toep_cholesky_lower_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  toep_cholesky_lower computes the lower Cholesky' )
  print ( '  factor of a positive semi-definite Toeplitz matrix.' )
  print ( '  The Toeplitz matrix is supplied as a (2,N) array.' )

  n = 3
  g = np.array ( [ \
    [ 1.0,   0.5, -0.375 ], \
    [ 0.0,   0.5, -0.375 ] ] )

  r8mat_print ( 2, n, g, '  Compressed Toeplitz matrix G:' )

  l = toep_cholesky_lower ( n, g )
  r8mat_print ( n, n, l, '  Computed lower Cholesky factor L:' )

  b = np.dot ( l, l.transpose ( ) )
  r8mat_print ( n, n, b, '  Product L*L\':' )

  return

def toep_cholesky_upper ( n, g ):

#*****************************************************************************80
#
## toep_cholesky_upper: upper Cholesky factor of a compressed Toeplitz matrix.
#
#  Discussion:
#
#    The Toeplitz matrix A is supplied in a compressed form G.
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = R' * R.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real G(2,N), the compressed Toeplitz matrix.
#    G(1,1:N) contains the first row.
#    G(2,2:N) contains the first column.
#
#  Output:
#
#    real R(N,N), the upper Cholesky factor.
#
  import numpy as np

  r = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    r[0,j] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      r[i,j] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return r

def toep_cholesky_upper_test ( ):

#*****************************************************************************80
#
## toep_cholesky_upper_test tests toep_cholesky_upper.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'toep_cholesky_upper_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  toep_cholesky_upper computes the upper Cholesky' )
  print ( '  factor of a positive semi-definite Toeplitz matrix.' )
  print ( '  The Toeplitz matrix is supplied as a (2,N) array.' )

  n = 3
  g = np.array ( [ \
    [ 1.0,   0.5, -0.375 ], \
    [ 0.0,   0.5, -0.375 ] ] )

  r8mat_print ( 2, n, g, '  Compressed Toeplitz matrix G:' )

  r = toep_cholesky_upper ( n, g )
  r8mat_print ( n, n, r, '  Computed upper Cholesky factor R:' )

  b = np.dot ( r.transpose ( ), r )
  r8mat_print ( n, n, b, '  Product R\'R:' )

  return

def toeplitz_cholesky_lower ( n, a ):

#*****************************************************************************80
#
## toeplitz_cholesky_lower computes the lower Cholesky factor of a Toeplitz matrix.
#
#  Discussion:
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = L * L'.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real A(N,N), the Toeplitz matrix.
#
#  Output:
#
#    real L(N,N), the lower Cholesky factor.
#
  import numpy as np

  g = np.zeros ( [ 2, n ] )

  for j in range ( 0, n ):
    g[0,j] = a[j,0]
  for j in range ( 1, n ):
    g[1,j] = a[j,0]

  l = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    l[j,0] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      l[j,i] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return l

def toeplitz_cholesky_lower_test ( ):

#*****************************************************************************80
#
## toeplitz_cholesky_lower_test tests toeplitz_cholesky_lower.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'toeplitz_cholesky_lower_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  toeplitz_cholesky_lower computes the lower Cholesky' )
  print ( '  factorization of a positive semi-definite Toeplitz matrix.' )
  print ( '  The matrix is supplied as an NxN array.' )

  n = 3
  a = np.array ( [ \
    [  1.0,   0.5, -0.375 ], \
    [  0.5,   1.0,  0.5   ], \
    [ -0.375, 0.5,  1.0   ] ] )

  r8mat_print ( n, n, a, '  Toeplitz matrix A:' )

  l = toeplitz_cholesky_lower ( n, a )
  r8mat_print ( n, n, l, '  Computed lower Cholesky factor L:' )

  b = np.dot ( l, l.transpose ( ) )
  r8mat_print ( n, n, b, '  Product LL\':' )

  return

def toeplitz_cholesky_upper ( n, a ):

#*****************************************************************************80
#
## toeplitz_cholesky_upper computes the upper Cholesky factor of a Toeplitz matrix.
#
#  Discussion:
#
#    The Toeplitz matrix must be positive semi-definite.
#
#    After factorization, A = R' * R.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    28 January 2017
#
#  Author:
#
#    Python version by John Burkardt
#
#  Reference:
#
#    Michael Stewart,
#    Cholesky factorization of semi-definite Toeplitz matrices.
#    Linear Algebra and its Applications,
#    Volume 254, pages 497-525, 1997.
#
#  Input:
#
#    integer N, the order of the matrix.
#
#    real A(N,N), the Toeplitz matrix.
#
#  Output:
#
#    real R(N,N), the upper Cholesky factor.
#
  import numpy as np

  g = np.zeros ( [ 2, n ] )

  for j in range ( 0, n ):
    g[0,j] = a[0,j]
  for j in range ( 1, n ):
    g[1,j] = a[j,0]

  r = np.zeros ( [ n, n ] )

  for j in range ( 0, n ):
    r[0,j] = g[0,j]

  for j in range ( n - 1, 0, -1 ):
    g[0,j] = g[0,j-1]
  g[0,0] = 0.0

  for i in range ( 1, n ):
    rho = - g[1,i] / g[0,i]
    gam = np.sqrt ( ( 1.0 - rho ) * ( 1.0 + rho ) )
    for j in range ( i, n ):
      alf = g[0,j]
      bet = g[1,j]
      g[0,j] = (       alf + rho * bet ) / gam
      g[1,j] = ( rho * alf +       bet ) / gam
    for j in range ( i, n ):
      r[i,j] = g[0,j]
    for j in range ( n - 1, i, -1 ):
      g[0,j] = g[0,j-1]
    g[0,i] = 0.0

  return r

def toeplitz_cholesky_upper_test ( ):

#*****************************************************************************80
#
## toeplitz_cholesky_upper_test tests toeplitz_cholesky_upper.
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    27 January 2017
#
#  Author:
#
#    John Burkardt
#
  import numpy as np
  import platform

  print ( '' )
  print ( 'toeplitz_cholesky_upper_test' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  toeplitz_cholesky_upper computes the upper Cholesky' )
  print ( '  factorization of a positive semi-definite Toeplitz matrix.' )
  print ( '  The matrix is supplied as an NxN array.' )

  n = 3
  a = np.array ( [ \
    [  1.0,   0.5, -0.375 ], \
    [  0.5,   1.0,  0.5   ], \
    [ -0.375, 0.5,  1.0   ] ] )

  r8mat_print ( n, n, a, '  Toeplitz matrix A:' )

  r = toeplitz_cholesky_upper ( n, a )
  r8mat_print ( n, n, r, '  Computed upper Cholesky factor R:' )

  b = np.dot ( r.transpose ( ), r )
  r8mat_print ( n, n, b, '  Product R\'R:' )

  return

def toeplitz_cholesky_test ( ):

#*****************************************************************************80
#
## toeplitz_cholesky_test() tests toeplitz_cholesky().
#
#  Licensing:
#
#    This code is distributed under the GNU LGPL license.
#
#  Modified:
#
#    27 January 2017
#
#  Author:
#
#    John Burkardt
#
  import platform

  print ( '' )
  print ( 'toeplitz_cholesky_test():' )
  print ( '  Python version: %s' % ( platform.python_version ( ) ) )
  print ( '  Test toeplitz_cholesky().' )

  t_cholesky_lower_test ( )
  toep_cholesky_lower_test ( )
  toeplitz_cholesky_lower_test ( )

  t_cholesky_upper_test ( )
  toep_cholesky_upper_test ( )
  toeplitz_cholesky_upper_test ( )
#
#  Terminate.
#
  print ( '' )
  print ( 'toeplitz_cholesky_test():' )
  print ( '  Normal end of execution.' )
  print ( '' )
  return

if ( __name__ == '__main__' ):
  timestamp ( )
  toeplitz_cholesky_test ( )
  timestamp ( )




# lsqfitgp/examples/dft.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""
Constrain the discrete Fourier transform of a periodic process. Shows how
to use GP.addlintransf.
"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
from jax import numpy as jnp
import gvar

class PORCO: pass
class DUO: pass

gp = lgp.GP(lgp.Zeta(scale=2 * np.pi, nu=0.5))
# We could do the same with a non-periodic prior, but it would not make as much
# sense. Moreover we use the Zeta kernel with nu=0.5, i.e., prior on the
# coefficients going down like 1/k. A smoother kernel would yield a strongly
# shrinking prior on the DFT coefficients. Try it.

xdata = np.linspace(0, 2 * np.pi, 20)
gp.addx(xdata, PORCO)

xpred = np.linspace(-2 * np.pi, 4 * np.pi, 50 * 6 + 1)
# If we don't align the number of points to the period, the samples will not
# show as periodic because the function is very rough. The distribution would
# still be periodic though, it's just an artifact of looking at a coarse grid.
gp.addx(xpred, DUO)

def transf(x):
    # use jax.numpy instead of numpy in this function
    f = jnp.fft.rfft(x)
    return jnp.stack([jnp.real(f), jnp.imag(f)])

gp.addlintransf(transf, [PORCO], 'dft')

gp.addlintransf(lambda x: x[0, 3], ['dft'], '3rd real coef')
gp.addlintransf(lambda x: x[0, 3] - x[1, 3], ['dft'], '3rd coef pseudo-phase')
# We will force the 3rd spectrum coefficient to be high because the supervisor
# said that it always comes out high, so the plot can't be different from the
# other articles. The statistics professor said that the correct fitting method
# is "Bayesian statistics" and that everything is subjective, so we may as
# well hardcode the desiratum into the prior.

u = gp.predfromdata({
    '3rd real coef': gvar.gvar(10, 1),
    '3rd coef pseudo-phase': gvar.gvar(5, 1),
})
mean = gvar.mean(u)
sdev = gvar.sdev(u)
cov = gvar.evalcov(u)

fig, axs = plt.subplots(2, 1, num='dft', clear=True, figsize=[6.4, 7])

ax = axs[0]
ax.set_title('Function')
m = mean[DUO]
s = sdev[DUO]
patch = ax.fill_between(xpred, m - s, m + s, alpha=0.5)
color = patch.get_facecolor()[0]
simulated_lines = np.random.multivariate_normal(m, cov[DUO, DUO])
ax.plot(xpred, simulated_lines, '-', color=color)
ax.plot(xdata, np.zeros_like(xdata), '.k', label='discrete lattice')

ax = axs[1]
ax.set_title('DFT')
simul = gvar.sample(u['dft'])
for i, label in enumerate(['real', 'imag']):
    m = mean['dft'][i]
    s = sdev['dft'][i]
    n = len(m)
    patch = ax.fill_between(np.arange(n), m - s, m + s, alpha=0.5, label=label)
    color = patch.get_facecolor()[0]
    ax.plot(np.arange(n), simul[i], color=color)

for ax in axs:
    ax.legend()
    ax.grid(linestyle='--')

fig.show()



import functools

import jax
import jax.numpy as jnp

@functools.singledispatch
def f(_):
    print('failure')

@f.register
def _(_: jnp.ndarray):
    print('success')

f(jnp.zeros(1)) # -> success

def g(x):
    f(x)
    return x

jax.jacobian(g)(jnp.zeros(1)) # -> failure
jax.jit(g)(jnp.zeros(1))
jax.jit(jax.jacobian(g))(jnp.zeros(1))

# class MyTracer(jax.core.Tracer):
#     __init__ = lambda _: None
#     aval = property(lambda _: jax.core.UnshapedArray(object))
#
# print(isinstance(MyTracer(), jnp.ndarray)) # -> True
# print(issubclass(jax.core.Tracer, jnp.ndarray)) # -> False
#
# ArrayMeta = type(jnp.ndarray)
# def __subclasscheck__(self, subclass):
#     if issubclass(subclass, jax.core.Tracer):
#         return True
#     else:
#         return super(ArrayMeta, self).__subclasscheck__(subclass)
# ArrayMeta.__subclasscheck__ = __subclasscheck__
# print(issubclass(jax.core.Tracer, jnp.ndarray)) # -> True
#
# jax.jacobian(g)(jnp.zeros(1)) # -> failure
#
# @f.register
# def _(_: jax.core.Tracer):
#     print('~success')
#
# jax.jacobian(g)(jnp.zeros(1)) # -> ~success
